#!/usr/bin/env python
# filename: phylogeny.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import colorsys
from collections import Counter
from copy import copy, deepcopy
import itertools
import math
import multiprocessing as mp
import os
import random
import shutil
import string
import subprocess as sp
import sys

import numpy as np
import scipy
import pandas as pd

from scipy.ndimage import gaussian_filter1d
from scipy.special import rel_entr

import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import seaborn as sns

import ete3

from Bio import AlignIO, Phylo

import abstar
from abstar.core.germline import get_imgt_germlines

from ..core.pair import Pair
from ..core.sequence import Sequence
from .alignment import mafft, muscle
from .cluster import cluster
from .color import hex_to_rgb, get_cmap
from .decorators import lazy_property

if sys.version_info[0] > 2:
    STR_TYPES = [str, ]
else:
    STR_TYPES = [str, unicode]




#--------------------------------
#  PHYLOGENETIC RECONSTRUCTION
#--------------------------------


def fasttree(alignment, tree_file, is_aa=False, quiet=True):
    if is_aa:
        ft_cmd = 'fasttree {} > {}'.format(alignment, tree_file)
    else:
        ft_cmd = 'fasttree -nt {} > {}'.format(alignment, tree_file)
    ft = sp.Popen(ft_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = ft.communicate()
    if not quiet:
        print(ft_cmd)
        print(stdout)
        print(stderr)
    return tree_file


def lsd(tree, output_file=None, dates_file=None, outgroup_file=None,
        with_constraints=True, with_weights=True, reestimate_root_position=None, quiet=True):
    lsd_cmd = 'lsd -i {}'.format(os.path.abspath(tree))
    if output_file is not None:
        lsd_cmd += ' -o {}'.format(os.path.abspath(output_file))
    if dates_file is not None:
        lsd_cmd += ' -d {}'.format(os.path.abspath(dates_file))
    if outgroup_file is not None:
        lsd_cmd += ' -g {}'.format(os.path.abspath(outgroup_file))
    if with_constraints:
        lsd_cmd += ' -c'
    if with_weights:
        lsd_cmd += ' -v'
    if reestimate_root_position is not None:
        lsd_cmd += ' -r {}'.format(reestimate_root_position)
    p = sp.Popen(lsd_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    if not quiet:
        print(lsd_cmd)
        print(stdout)
        print(stderr)
    return output_file


def igphyml(input_file=None, tree_file=None, root=None, verbose=False):
    '''
    Computes a phylogenetic tree using IgPhyML.

    .. note::
        
        IgPhyML must be installed. It can be downloaded from https://github.com/kbhoehn/IgPhyML.

    Args:

        input_file (str): Path to a Phylip-formatted multiple sequence alignment. Required.

        tree_file (str): Path to the output tree file.

        root (str): Name of the root sequence. Required.

        verbose (bool): If `True`, prints the standard output and standard error for each IgPhyML run. 
            Default is `False`.
    '''

    if shutil.which('igphyml') is None:
        raise RuntimeError('It appears that IgPhyML is not installed.\nPlease install and try again.')
    
    # first, tree topology is estimated with the M0/GY94 model
    igphyml_cmd1 = 'igphyml -i {} -m GY -w M0 -t e --run_id gy94'.format(aln_file)
    p1 = sp.Popen(igphyml_cmd1, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout1, stderr1 = p1.communicate()
    if verbose:
        print(stdout1 + '\n')
        print(stderr1 + '\n\n')
    intermediate = input_file + '_igphyml_tree.txt_gy94'

    # now  we fit the HLP17 model once the tree topology is fixed
    igphyml_cmd2 = 'igphyml -i {0} -m HLP17 --root {1} -o lr -u {}_igphyml_tree.txt_gy94 -o {}'.format(input_file,
                                                                                                       root,
                                                                                                       tree_file)
    p2 = sp.Popen(igphyml_cmd2, stdout=sp.PIPE, stderr=sp.PIPE)
    stdout2, stderr2 = p2.communicate()
    if verbose:
        print(stdout2 + '\n')
        print(stderr2 + '\n')
    return tree_file + '_igphyml_tree.txt'




#--------------------------------
#      PHYLOGENETIC TREES
#--------------------------------


def phylogeny(sequences=None, project_dir=None, name=None, aln_file=None, tree_file=None,
        seq_field=None, name_field=None, aa=False, species='human', unrooted=False, ladderize=True,
        root=None, root_name=None, show_root_name=False, color_dict=None, color_function=None,
        order_dict=None, order_function=None, color_node_labels=False, label_colors=None,
        scale=None, branch_vert_margin=None, fontsize=12, show_names=True, show_scale=False,
        mirror=False, min_order_fraction=0.1, figname_prefix=None, figname_suffix=None,
        # linked_alignment=None, alignment_fontsize=11, alignment_height=50, alignment_width=50, 
        compact_alignment=False, scale_factor=1, rename_function=None, linewidth=1.0,
        delete_nodes=None, quiet=True):
    '''
    Generates a lineage phylogeny figure.

    Args:

        sequences (list(Sequence)): A list of ``Sequence`` objects from which a phylogeny
            will be calculated. Strictly speaking, they do not need to be ``Sequence`` objects,
            rather, any object that contains the sequence name as the ``id`` attribute (or
            by dictionary-style lookup using the provided ``name_field``) and contains the
            sequence as the ``sequence`` attribute (or by dictionary-stype lookup using the
            provided ``seq_field``).

        project_dir (str): directory into which all phylogeny files will be deposited,
            including alignment, tree and figure files.

        name (str): Name to be used for naming alignment, tree, and phylogeny files. If not
            provided, a random name will be generated.

        aln_file (str): if a multiple sequence alignment has already been calculated,
            passing the path to the alignment file (in FASTA format) will force Lineage.phylogeny()
            to use the supplied msa instead of computing a new one.

        tree_file (str): if a tree file has already been calculated, passing the path
            to the pre-computed tree file will force ``phylogeny()`` to use
            the supplied tree file instead of computing a new one. It is important to note that
            only sequence names will be parsed from the tree_file, so if ``order_function`` or
            ``color_function`` is also provided, ensure that these functions only require the
            sequence ID rather than the entire sequence.

        aa (bool): if True, use amino acid sequences to compute the phylogeny.
            Default is False.

        root (Sequence, str: The root can be provided either as a ``Sequence`` object (if ``sequences``
            are being provided) or as the name of a sequence that can be found either in
            ``sequences`` or in the provided ``aln_file`` or ``tree_file``. Note that if
            either ``aln_file`` or ``tree_file`` are provided, the root must be provided
            as the sequence name, not as a ``Sequence`` object (as the root sequence must
            already be included in either ``aln_file`` or ``tree_file``. If the root is not
            provided, the germline V-gene sequence of the

        color_dict (dict): Dictionary with sequence IDs as keys and colors (hex format) as values. If any
            sequence IDs are not found in the dict, they will be colored black. If neither ``color_dict`` nor
            ``color_function`` is provided, all leaves will be colored black.

        color_function (func): Function that that accepts a ``Sequence`` object and returns the color
            (as a hex code). If ``color_dict`` is also provided, ``color_function`` is ignored. Additionally,
            ``color_function`` will only be used if ``sequences`` are provided. If ``sequences`` are not provided
            (instead using ``aln_file` or ``tree_file``), ``color_dict`` must be used instead of ``color_function``.

        orders: a dictionary with sequence IDs as keys and orders (integers) as values.
            If not provided, only the leaf branches will be colored (if <colors> or
            <color_function> is provided).

        chain: build a phylogeny using the given chain ('heavy' or 'light').
            Default is 'heavy'.

        filter_function: function used to filter sequences (identity-based clustering, for
            example). The function should accept a list of Sequence objects and return
            a list of Sequence objects.

        just_pairs: if True, compute the phylogeny using only paired sequences.
            Default (False) will use all sequences of the appropriate chain, paired or not.

        scale (float): passed to ete3.TreeStyle() to set the scale of the tree figure. Increased
            scale results in a wider tree.

        branch_vert_margin (int): passed to ete3.TreeStyle() to set the branch_vertical_margin of
            the tree figure. Increased branch_vert_margin results in a taller tree.

        fontsize: size of the leaf labels. Default is 12.

        show_names: show names of leaf nodes. Options are True (show labels for all leaf nodes),
            False (don't show labels for any leaf nodes) or a list of sequence IDs for which
            labels should be shown. Default is True.

        mirror: flip the orientation of the tree. Default is to draw the tree from left to right.
            Setting mirror to True results in the tree being drawn from right to left.

        min_order_fraction: minimum fraction of downstream leaves requried to color a branch.
            When coloring non-leaf nodes, the earliest 'order' with at least <min_order_fraction>
            leaf nodes is used. Default is 0.1 (which corresponds to 10%).

        figname_prefix: by default, figures will be named <lineage_id>.pdf. If prefix='prefix_' and
            the lineage ID is 'ABC123', the figure file will be named 'prefix_ABC123.pdf'.

        figname_suffix: by default, figures will be named <lineage_id>.pdf. If suffix='_suffix' and
            the lineage ID is 'ABC123', the figure file will be named 'ABC123_suffix.pdf'.
    '''

    if project_dir is None:
        print('\nERROR: project_dir is required\n')
        sys.exit(1)
    else:
        project_dir = os.path.abspath(project_dir)

    # make a name if one isn't provided
    if name is None:
        name = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))

    # if sequences are provided, need to process them
    if sequences is not None and all([arg is None for arg in [aln_file, tree_file]]):
        sequences = deepcopy(sequences)
        root = copy(root)

        # if custom seq_field is specified, copy to the .sequence attribute
        if seq_field is not None:
            if not all([seq_field in list(s.annotations.keys()) for s in sequences]):
                print('\nERROR: {} is not present in all of the supplied sequences.\n'.format(seq_field))
                sys.exit(1)
            for s in sequences:
                s.alignment_sequence = s[seq_field]
        else:
            for s in sequences:
                s.alignment_sequence = s.sequence

        # if custom name_field is specified, copy to the .id attribute
        if name_field is not None:
            if not all([name_field in list(s.annotations.keys()) for s in sequences]):
                print('\nERROR: {} is not present in all of the supplied sequences.\n'.format(name_field))
                sys.exit(1)
            for s in sequences:
                s.alignment_id = s[name_field]
        else:
            for s in sequences:
                s.alignment_id = s.id

        # parse the root sequence
        if unrooted:
            root = None
            root_name = None
        elif root is None:
            if not quiet:
                print('\nRoot sequence was was not provided. Using the germline V-gene.')
            if not all(['v_gene' in list(s.annotations.keys()) for s in sequences]):
                print('\nInput sequences to not appear to be AbStar annotated. Annotating now...')
                sequences = abstar.run(*[(s.id, s.sequence) for s in sequences])
                print('Done.')
            if not all(['full' in list(s['v_gene'].keys()) for s in sequences]):
                print('\nInput sequences to not appear to be AbStar annotated. Annotating now...')
                sequences = abstar.run(*[(s.id, s.sequence) for s in sequences])
                print('Done.')
            top_vgene = sorted(list(Counter([s['v_gene']['full'] for s in sequences]).items()),
                               key=lambda x: x[1],
                               reverse=True)[0][0]
            vgene = get_imgt_germlines(species, 'V', gene=top_vgene)
            if aa:
                root = Sequence(vgene.ungapped_aa_sequence, id=top_vgene)
            else:
                root = Sequence(vgene.ungapped_nt_sequence, id=top_vgene)
            root.alignment_id = root.id
            root.alignment_sequence = root.sequence
            if not quiet:
                print('Top V-gene: {}'.format(root.alignment_id))
                print(root.alignment_sequence)
        elif type(root) in STR_TYPES:
            root = [s for s in sequences if s.alignment_id == root][0]
            if not root:
                print('\nERROR: The name of the root sequence ({}) was not found in the list of input sequences.'.format(root))
                print('\n')
                sys.exit(1)
            sequences = [s for s in sequences if s.alignment_id != root.alignment_id]
        elif type(root) == Sequence:
            if seq_field is not None:
                if seq_field not in list(root.anotations.keys()):
                    print('\nERROR: {} is not present in the supplied root sequence.\n'.format(seq_field))
                    sys.exit(1)
                root.alignment_sequence = root[seq_field]
            if name_field is not None:
                if name_field not in list(root.anotations.keys()):
                    print('\nERROR: {} is not present in the supplied root sequence.\n'.format(name_field))
                    sys.exit(1)
                root.alignment_id = root[name_field]
            sequences = [s for s in sequences if s.alignment_id != root.alignment_id]
        else:
            print('\nERROR: If root is provided, it must be the name of a sequence \
            found in the supplied list of sequences or it must be a Sequence object.')
            print('\n')
            sys.exit(1)
        if not unrooted:
            if root_name is not None:
                root.alignment_id = root_name
            else:
                root_name = root.alignment_id
            sequences.append(root)

    # parse sequences from aln_file, if provided
    elif aln_file is not None:
        if not unrooted and type(root) not in STR_TYPES:
            print('\nERROR: If providing an aln_file, the name of the root sequence must \
            be provided (as a string) using the root keyword argument')
            print('\n')
            sys.exit(1)
        _sequences = []
        _root = None
        for rec in AlignIO.read(open(aln_file), 'fasta'):
            s = str(rec.seq).replace('-', '')
            if rec.id == root:
                _root = Sequence(s, rec.id)
                _root.alignment_id = _root.id
            else:
                _s = Sequence(s, id=rec.id)
                _s.alignment_id = rec.id
                _sequences.append(_s)
        if sequences is None:
            sequences = _sequences
        else:
            sequence_ids = [s.id for s in sequences]
            if any([_s.alignment_id not in sequence_ids for _s in _sequences]):
                print('\nWARNING: Sequences were found in the alignment file that were not included \
                      in the input sequence list. This may cause problems.')
            for s in sequences:
                s.alignment_id = s.id
                s.alignment_sequence = s.sequence
        if unrooted:
            root = None
            root_name = None
        else:
            if _root is None:
                print('\nERROR: The specified root ({}) was not found in the provided alignment file.'.format(root))
                print('\n')
                sys.exit(1)
            root = _root
            if root_name is not None:
                root.alignment_id = root_name
            else:
                root_name = root.alignment_id
            sequences = [s for s in sequences if all([s.alignment_id != name for name in [root.id, root.alignment_id]])]
            sequences.append(root)

    # parse sequences from tree_file, if provided
    elif tree_file is not None:
        if not unrooted and type(root) not in STR_TYPES:
            print('\nERROR: If providing a tree_file, the name of the root sequence must \
            be provided (as a string) using the root keyword argument')
            print('\n')
            sys.exit(1)
        _sequences = []
        _root = None
        tree = Phylo.read(open(tree_file), 'newick')
        for leaf in tree.get_terminals():
            s = ''
            if leaf.name == root:
                _root = Sequence(s, leaf.name)
                _root.alignment_id = _root.id
            else:
                _s = Sequence(s, id=leaf.name)
                _s.alignment_id = leaf.name
                _sequences.append(_s)
        if sequences is None:
            sequences = _sequences
        else:
            sequence_ids = [s.id for s in sequences]
            if any([_s.alignment_id not in sequence_ids for _s in _sequences]):
                print('\nWARNING: Sequences were found in the alignment file that were not included \
                      in the input sequence list. This may cause problems.')
            for s in sequences:
                s.alignment_id = s.id
                s.alignment_sequence = s.sequence
        if unrooted:
            root = None
            root_name = None
        elif _root is None:
            print('\nERROR: The specified root ({}) was not found in the provided tree file.'.format(root))
            print('\n')
            sys.exit(1)
        else:
            root = _root
            if root_name is not None:
                root.alignment_id = root_name
            else:
                root_name = root.alignment_id
            sequences = [s for s in sequences if all([s.alignment_id != name for name in [root.id, root.alignment_id]])]
            sequences.append(root)

    # set up colors and color ordering
    if order_dict is None:
        if order_function is not None:
            order_dict = {seq.alignment_id: order_function(seq) for seq in sequences}
    if color_dict is None:
        if color_function is not None:
            color_dict = {seq.alignment_id: color_function(seq) for seq in sequences}
    if color_dict is None:
        color_dict = {}

    # make msa (if necessary)
    if all([aln_file is None, tree_file is None]):
        aln_file = os.path.abspath(os.path.join(project_dir, '{}.aln'.format(name)))
        # muscle(seqs, aln_file, as_file=True)
        do_print = False if quiet else True
        if do_print:
            print('\n')
        seqs = [(s.alignment_id, s.alignment_sequence) for s in sequences]
        mafft(seqs, aln_file, as_file=True, print_stdout=do_print, print_stderr=do_print)

    # make treefile (if necessary)
    if tree_file is None:
        tree_file = os.path.abspath(os.path.join(project_dir, '{}.nw'.format(name)))
        fasttree(aln_file, tree_file, is_aa=aa, quiet=quiet)

    # make phylogeny
    prefix = '' if figname_prefix is None else figname_prefix
    suffix = '' if figname_suffix is None else figname_suffix
    fig_file = os.path.join(project_dir, '{}{}{}.pdf'.format(prefix, name, suffix))
    _make_tree_figure(tree_file,
                      fig_file,
                      color_dict,
                      order_dict,
                      None if root is None else root.alignment_id,
                      rename_function=rename_function,
                      show_names=show_names,
                      name_field=name_field,
                      branch_vert_margin=branch_vert_margin,
                      scale=scale,
                      color_node_labels=color_node_labels,
                      label_colors=label_colors,
                      show_root_name=show_root_name,
                      tree_orientation=1 if mirror else 0,
                      fontsize=fontsize,
                      min_order_fraction=min_order_fraction,
                    #   linked_alignment=linked_alignment,
                    #   alignment_fontsize=alignment_fontsize,
                    #   alignment_height=alignment_height,
                    #   alignment_width=alignment_width,
                      show_scale=show_scale,
                      compact_alignment=compact_alignment,
                      scale_factor=scale_factor,
                      linewidth=linewidth,
                      ladderize=ladderize,
                      delete_nodes=delete_nodes)


def _make_tree_figure(tree, fig, colors, orders, root_name, scale=None, branch_vert_margin=None,
        fontsize=12, show_names=True, name_field='seq_id', rename_function=None, color_node_labels=False, label_colors=None,
        tree_orientation=0, min_order_fraction=0.1, show_root_name=False, chain=None,
        # linked_alignment=None, alignment_fontsize=11, alignment_height=50, alignment_width=50,
        compact_alignment=False, scale_factor=1, linewidth=1, show_scale=False, ladderize=True, delete_nodes=None):
    if delete_nodes is None:
        delete_nodes = []
    elif type(delete_nodes) in STR_TYPES:
        delete_nodes = [delete_nodes, ]
    if show_root_name is True:
        show_names.append(root_name)
    # if linked_alignment is not None:
    #     t = ete3.PhyloTree(tree, alignment=linked_alignment, alg_format='fasta')
    #     ete3.faces.SequenceItem = MySequenceItem
    t = ete3.Tree(tree)
    if root_name is not None:
        t.set_outgroup(t&root_name)
    # style the nodes
    for node in t.traverse():
        if node.name in delete_nodes:
            node.delete()
            continue
        if orders is not None:
            leaves = node.get_leaf_names()
            order_count = Counter([orders[l] for l in leaves])
            for order in sorted(order_count.keys()):
                if float(order_count[order]) / len(leaves) >= min_order_fraction:
                    color = colors[order]
                    break
        else:
            color = colors.get(node.name, '#000000')
        # if linked_alignment is not None:
        #     node.add_feature('aln_fontsize', alignment_fontsize)
        #     node.add_feature('aln_height', alignment_height)
        #     node.add_feature('aln_width', alignment_width)
        #     node.add_feature('fontsize', fontsize)
        #     node.add_feature('format', 'seq')
        #     node.add_feature('scale_factor', scale_factor)
        style = ete3.NodeStyle()
        style['size'] = 0
        style['vt_line_width'] = float(linewidth)
        style['hz_line_width'] = float(linewidth)
        style['vt_line_color'] = color
        style['hz_line_color'] = color
        style['vt_line_type'] = 0
        style['hz_line_type'] = 0
        if show_names is True:
            tf = _build_node_text_face(node, color_node_labels, color, label_colors, fontsize, rename_function)
            node.add_face(tf, column=0)
        elif node.name in show_names:
            tf = _build_node_text_face(node, color_node_labels, color, label_colors, fontsize, rename_function)
            node.add_face(tf, column=0)
        node.set_style(style)
    t.dist = 0
    ts = ete3.TreeStyle()
    # if linked_alignment is not None:
    #     ts.layout_fn = _phyloalignment_layout_function
    ts.orientation = tree_orientation
    ts.show_leaf_name = False
    if scale is not None:
        ts.scale = int(scale)
    if branch_vert_margin is not None:
        ts.branch_vertical_margin = float(branch_vert_margin)
    ts.show_scale = show_scale
    if ladderize:
        t.ladderize()
    t.render(fig, tree_style=ts)


def _build_node_text_face(node, color_node_labels, color, label_colors, fontsize, rename_function):
    if color_node_labels:
        if label_colors is None:
            node_color = color
        elif type(label_colors) == dict:
            node_color = label_colors.get(node.name, '#000000')
        elif type(label_colors) in [list, tuple]:
            node_color = color if node.name in label_colors else '#000000'
        else:
            node_color = '#000000'
    else:
        node_color = '#000000'
    node_name = node.name if rename_function is None else rename_function(node.name)
    tf = ete3.TextFace(node_name,
                       fsize=fontsize,
                       fgcolor=node_color)
    return tf




#-----------------------------------------
#      PHYLOGENETIC LAPLACIAN SPECTRA
#-----------------------------------------


class PLS():
    '''
    Object representing a collection of Phylogenetic Laplacian Spectra.
    '''
    def __init__(self, spectral_trees, n_processes=0):
        self.trees = spectral_trees
        self.n_processes = mp.cpu_count() if n_processes == 0 else n_processes
        self.compute_kdes()

    
    @lazy_property
    def distance_matrix(self):
        dist = {}
        p = mp.Pool(processes=self.n_processes)
        async_results = []
        for t in self.trees:
            async_results.append(p.apply_async(self.calc_distance_row, args=(t, self.trees)))
        dists = [ar.get() for ar in async_results]
        p.close()
        p.join()
        for t1, row in zip(self.trees, dists):
            for t2, d in zip(self.trees, row):
                dist[t1.name][t2.name] = d
        return pd.DataFrame(dist)


        # for t1, t2 in itertools.product(self.trees, repeat=2):
        #     dist[t1.name] = {}
        #     if t1.name not in dist:
        #         dist[t1.name] = {}
        #     dist[t1.name][t2.name] = self.compare_trees(t1, t2)
        # return pd.DataFrame(dist)


    def calc_distance_row(self, t1, trees):
        dists = []
        for t2 in trees:
            dists.append(self.compare_trees(t1, t2))
        return dists


    def compare_trees(self, t1, t2, n_points=100):
        maxw = max([t1.principal_eigenvalue, t2.principal_eigenvalue])
        x = np.linspace(0, maxw, n_points)
        p = t1.pdf(x)
        q = t2.pdf(x)
        return self.jensenshannon(p, q)

    @staticmethod
    def jensenshannon(p, q, base=None):
        p = np.asarray(p)
        q = np.asarray(q)
        p = p / np.sum(p, axis=0)
        q = q / np.sum(q, axis=0)
        m = (p + q) / 2.0
        left = rel_entr(p, m)
        right = rel_entr(q, m)
        js = np.sum(left, axis=0) + np.sum(right, axis=0)
        if base is not None:
            js /= np.log(base)
        return np.sqrt(js / 2.0)


    def compute_kdes(self):
        if self.n_processes == 1:
            self._compute_kdes_sp()
        else:
            self._compute_kdes_mp()

    def _compute_kdes_sp(self):
        for t in self.trees:
            kde = t.kde

    def _compute_kdes_mp(self):
        p = mp.Pool(processes=self.n_processes)
        async_results = []
        for t in self.trees:
            async_results.append(p.apply_async(self.compute_kde, args=(t,)))
        kdes = [ar.get() for ar in async_results]
        for t, kde in zip(self.trees, kdes):
            t.kde = kde
        p.close()
        p.join()

    @staticmethod
    def compute_kde(tree):
        '''
        For a SpectralTree, computes the KDE.

        Args:

            tree (SpectralTree): tree for which the KDE will be computed

        Returns:

            the KDE function
        '''
        return tree.kde

    
    def clustermap(self, cmap=None, figfile=None):
        cm = sns.clustermap(self.distance_matrix, cmap=cmap)
        if figfile is not None:
            cm.savefig(figfile)
        else:
            plt.show()


class SpectralTree():
    
    def __init__(self, tree_file, subject=None):
        self.subject = subject
        self.name = os.path.basename(tree_file)
        self.tree_file = tree_file
        self.t = ete3.Tree(tree_file)
        
    
    @lazy_property
    def nodes(self):
        num = 0
        nodes = []
        for n in self.t.traverse():
            if not n.name:
                n.name = str(num)
                num += 1
            nodes.append(n)
        return nodes
    
    @lazy_property
    def node_names(self):
        return [n.name for n in self.nodes]
    
    @lazy_property
    def distances(self):
        distances = []
        for n1 in self.nodes:
            dist = {}
            for n2 in self.nodes:
                if n1 == n2:
                    continue
                dist[n2] = self.t.get_distance(n1, n2)
            dist[n1] = -1. * sum(dist.values())
            distances.append([dist[n] for n in self.nodes])
        return np.asarray(distances)
    
    @lazy_property
    def eigenvalues(self):
        gl = scipy.sparse.csgraph.laplacian(self.distances)
        w, v = np.linalg.eig(gl)
        return [e for e in sorted(w, reverse=True) if e >= 1]
    
    @lazy_property
    def log_eigenvalues(self):
        return np.real(np.log(self.eigenvalues))
    
    @property
    def principal_eigenvalue(self):
        return max(self.eigenvalues)

    @lazy_property
    def kde(self):
        return scipy.stats.gaussian_kde(self.eigenvalues)

    
    def pdf(self, x):
        return [v.real for v in self.kde.pdf(x)]
 

    def plot(self, xs=None, figfile=None, xlim=None):
        if xs is None:
            xs = np.linspace(-0.5, self.principal_eigenvalue, 200)
        ys = self.pdf(xs)
        # initialize and plot
        plt.figure(figsize=[3, 4])
        plt.plot(xs, ys, lw=2)
        plt.fill_between(xs, ys, alpha=0.2)
        # style
        ax = plt.gca()
        ax.set_xlabel('eigenvalue (ln)')
        ax.set_ylabel('density')
        if xlim is not None:
            ax.set_xlim(xlim)
        if figfile is not None:
            plt.tight_layout()
            plt.savefig(figfile)
        else:
            plt.show()



#!/usr/bin/env python
# filename: cluster.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime
import multiprocessing as mp
import os
import platform
import random
import shutil
import sqlite3
import string
import subprocess as sp
import sys
import tempfile
import time

from Bio import SeqIO, AlignIO
from Bio.Align import AlignInfo

from .alignment import mafft
from .database import KeyValueStore
from .decorators import lazy_property
from .pipeline import make_dir
from ..core.sequence import Sequence, read_fasta


class Cluster():
    '''
    Docstring for Cluster.
    '''
    def __init__(self, name, sequences, centroid=None):
        self.name = name
        self.sequences = sequences
        self.centroid = centroid
        
    @property
    def size(self):
        return len(self.sequences)
    
    @property
    def seq_ids(self):
        return [s.id for s in self.sequences]
    
    @lazy_property
    def consensus(self):
        return self._make_consensus()

    def _make_consensus(self):
        if len(self.sequences) == 1:
            return self.sequences[0]
        aln = mafft(self.sequences)
        if aln is None:
            print("ERROR: Failed to generate an alignmnet for a consensus sequence.")
            return None
        summary_align = AlignInfo.SummaryInfo(aln)
        consensus = summary_align.gap_consensus(threshold=0.51, ambiguous='n')
        consensus_string = str(consensus).replace('-', '')
        consensus_seq = Sequence(consensus_string.upper())
        return consensus_seq


class Clusters():
    '''
    Docstring for Clusters.
    '''
    def __init__(self, clusters=None):
        self._clusters = clusters if clusters is not None else []
        
    def __iter__(self):
        for cluster in self.clusters:
            yield cluster
            
    def __getitem__(self, key):
        return self.clusters[key]
    
    def __len__(self):
        return len(self._clusters)
    
    @property
    def clusters(self):
        return sorted(self._clusters, key=lambda x: x.size, reverse=True)
    
    @property
    def largest_cluster(self):
        return self.clusters[0]
            
    @property
    def count(self):
        return len(self.clusters)
    
    def add(self, cluster):
        self._clusters.append(cluster)



def cluster(sequences, threshold=0.975, temp_dir='/tmp', in_memory_db=True, debug=False):
    # preflight
    if isinstance(sequences, Sequence):
        sequences = [sequences, ]
    elif isinstance(sequences, (list, tuple)):
        pass
    elif os.path.isfile(sequences):
        sequences = read_fasta(sequences)
    ts = datetime.now().strftime("%Y-%m-%d_%H:%M:%S.%f")
    temp_dir = os.path.join(temp_dir, 'clustering_{}'.format(ts))
    make_dir(temp_dir)
    
    # set up a sequence db
    if in_memory_db:
        db = KeyValueStore(in_memory=True)
    else:
        db = KeyValueStore(name='mmseqs', direc=temp_dir)
    for s in sequences:
        db[s.id] = s
    db.index()
    
    # make the input FASTA file
    fasta_file = os.path.join(temp_dir, 'input.fasta')
    with open(fasta_file, 'w') as f:
        fstring = [s.fasta for s in sequences]
        f.write('\n'.join(fstring))
    
    # cluster
    clu_dict = mmseqs_cluster(fasta_file, min_seq_id=threshold, temp_dir=temp_dir, debug=debug)
    
    # parse clustering output
    clus = []
    for name, seqs in clu_dict.items():
        seq_ids = seqs['seq_ids']
        centroid_id = seqs['centroid']
        seqs = db.find_many(seq_ids)
        centroid = db.find_one(centroid_id)
        clus.append(Cluster(name, seqs, centroid=centroid))
    clusters = Clusters(clus)
    
    # clean up
    if not debug:
        db.destroy()
        shutil.rmtree(temp_dir)
    
    return clusters




def mmseqs_cluster(fasta_file, min_seq_id=0.975, temp_dir='/tmp', debug=False):
    # preflight
    fasta_file = os.path.abspath(fasta_file)
    temp_dir = os.path.join(temp_dir, 'mmseqs')
    make_dir(temp_dir)
    mod_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    mmseqs_bin = os.path.join(mod_dir, 'bin/mmseqs_{}'.format(platform.system().lower()))

    # create the sequence DB
    db_file = os.path.join(temp_dir, 'DB')
    db_cmd = '{} createdb {} {}'.format(mmseqs_bin, fasta_file, db_file)
    p = sp.Popen(db_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    if debug:
        if sys.version_info[0] > 2:
            stdout = stdout.decode('utf-8')
            stderr = stderr.decode('utf-8')
        print('STDOUT:', stdout)
        print('')
        print('STDERR:', stderr)

    # do the clustering
    clu_file = os.path.join(temp_dir, 'CLU')
    cluster_cmd = '{} cluster {} {} {} --min-seq-id {}'.format(mmseqs_bin, db_file, clu_file, temp_dir, min_seq_id)
    p = sp.Popen(cluster_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    if debug:
        if sys.version_info[0] > 2:
            stdout = stdout.decode('utf-8')
            stderr = stderr.decode('utf-8')
        print('STDOUT:', stdout)
        print('')
        print('STDERR:', stderr)

    # generate TSV-formatted output
    tsv_file = os.path.join(temp_dir, 'CLU.tsv')
    tsv_cmd = '{0} createtsv {1} {1} {2} {3}'.format(mmseqs_bin, db_file, clu_file, tsv_file)
    p = sp.Popen(tsv_cmd, stdout=sp.PIPE, stderr=sp.PIPE, shell=True)
    stdout, stderr = p.communicate()
    if debug:
        if sys.version_info[0] > 2:
            stdout = stdout.decode('utf-8')
            stderr = stderr.decode('utf-8')
        print('STDOUT:', stdout)
        print('')
        print('STDERR:', stderr)

    # parse TSV output
    clu_names = {}
    clu_seqs = {}
    with open(tsv_file) as f:
        for line in f:
            line = line.strip()
            if line.startswith('#'):
                continue
            c, s = line.split()
            if c not in clu_names:
                cname = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
                clu_names[c] = cname
            clu_name = clu_names[c]
            if clu_name not in clu_seqs:
                clu_seqs[clu_name] = {'centroid': c, 'seq_ids': []}
            clu_seqs[clu_name]['seq_ids'].append(s)

    # clean up
    if not debug:
        shutil.rmtree(temp_dir)
    return clu_seqs




# __all__ = ['CDHITResult', 'Cluster', 'cluster', 'cdhit', 'parse_clusters']


# class CDHITResult(object):
#     '''
#     Docstring for CDHITResult.
#     '''
#     def __init__(self, clusters, seq_db=None, db_path=None, seq_dict=None):
#         self.clusters = sorted(clusters, key = lambda x: x.size, reverse=True)
#         self.seq_db = seq_db
#         self.db_path = db_path
#         self.seq_dict = seq_dict
    
#     def __len__(self):
#         return len(self.clusters)

#     def __iter__(self):
#         for cluster in self.clusters:
#             yield cluster


#     @property
#     def largest_cluster(self):
#         return self.clusters[0]

    
#     def delete(self):
#         self.seq_db.connection.close()
#         os.unlink(self.db_path)


# class Cluster(object):
#     """
#     Data and methods for a cluster of sequences.

#     All public attributes are evaluated lazily, so attributes that
#     require significant processing time are only computed when needed.
#     In addition, attributes are cached after calculation, so if you
#     change the Cluster object after accessing attributes, the
#     attributes will not automatically update. Setters are provided for all attributes,
#     however, so you can update them manually if necessary::

#         seqs = [Sequence1, Sequence2, ... SequenceN]
#         clust = cluster(seqs)

#         # calculate the consensus
#         consensus = clust.consensus

#         # add sequences to the Cluster
#         more_sequences = [SequenceA, SequenceB, SequenceC]
#         clust.sequences += more_sequences

#         # need to recompute the consensus manually
#         clust.consensus = clust._make_consensus()


#     Attributes:

#         ids (list): A list of all sequence IDs in the Cluster

#         size (int): Number of sequences in the Cluster

#         sequences (list): A list of all sequences in the Cluster,
#             as AbTools ``Sequence`` objects.

#         consensus (Sequence): Consensus sequence, calculated by
#             aligning all sequences with MAFFT and computing the
#             ``Bio.Align.AlignInfo.SummaryInfo.gap_consensus()``

#         centroid (Sequence): Centroid sequence, as calculated by
#             CD-HIT.
#     """
#     def __init__(self, raw_cluster, seq_db=None, db_path=None, seq_dict=None):
#         super(Cluster, self).__init__()
#         self._raw_cluster = raw_cluster
#         self._seq_db = seq_db
#         self._seq_db_path = db_path
#         self._seq_dict = seq_dict


#     @lazy_property
#     def ids(self):
#         return self._get_ids()

#     @lazy_property
#     def size(self):
#         return len(self.ids)

#     @lazy_property
#     def sequences(self):
#         if all([self._seq_db is None, self._seq_dict is None]):
#             err = "ERROR: In order to access a Cluster's sequences, you must provide "
#             err += 'either a SQLite database connection object or a dictionary of sequences at instantiation.'
#             raise RuntimeError(err)
#         return self._get_sequences()

#     @lazy_property
#     def consensus(self):
#         if all([self._seq_db is None, self._seq_dict is None]):
#             err = "ERROR: In order to compute a Cluster's consensus, you must provide "
#             err += 'either a SQLite database connection object or a dictionary of sequences at instantiation.'
#             raise RuntimeError(err)
#         return self._make_consensus()

#     @lazy_property
#     def centroid(self):
#         if all([self._seq_db is None, self._seq_dict is None]):
#             err = "ERROR: In order to retrieve a Cluster's centroid, you must provide "
#             err += 'either a SQLite database connection object or a dictionary of sequences at instantiation.'
#             raise RuntimeError(err)
#         return self._get_centroid()


#     def cleanup(self):
#         try:
#             os.unlink(self._raw_cluster)
#         except:
#             pass
#         self.terminate_db()


#     def terminate_db(self):
#         if self._seq_db is not None:
#             self._seq_db.close()
#         if self._seq_db_path is not None:
#             if os.path.isfile(self._seq_db_path):
#                 os.unlink(self._seq_db_path)


#     def _get_ids(self):
#         ids = []
#         for c in self._raw_cluster[1:]:
#             if c:
#                 ids.append(c.split()[2][1:-3])
#         return ids

#     def _get_sequences(self):
#         if self._seq_db is not None:
#             seqs = []
#             for chunk in self._chunker(self.ids):
#                 sql_cmd = '''SELECT seqs.id, seqs.sequence
#                              FROM seqs
#                              WHERE seqs.id IN ({})'''.format(','.join('?' * len(chunk)))
#                 seq_chunk = self._seq_db.execute(sql_cmd, chunk)
#                 seqs.extend(seq_chunk)
#             return [Sequence(s) for s in seqs]
#         else:
#             return [self._seq_dict[s] for s in self.ids]

#     def _get_centroid(self):
#         for line in self._raw_cluster:
#            if '*' in line:
#                 centroid_id = line.split()[2][1:-3]
#                 break
#         centroid = [s for s in self.sequences if s.id == centroid_id][0]
#         return centroid

#     def _make_consensus(self):
#         if len(self.sequences) == 1:
#             return self.sequences[0]
#         _aln = mafft(self.sequences, as_file=True)
#         aln = AlignIO.read(open(_aln, 'r'), 'fasta')
#         summary_align = AlignInfo.SummaryInfo(aln)
#         consensus = summary_align.gap_consensus(threshold=0.51, ambiguous='n')
#         consensus_string = str(consensus).replace('-', '')
#         consensus_seq = Sequence(consensus_string.upper())
#         os.unlink(_aln)
#         return consensus_seq

#     @staticmethod
#     def _chunker(l, size=900):
#         return (l[pos:pos + size] for pos in range(0, len(l), size))


# def cluster(seqs, threshold=0.975, out_file=None, temp_dir=None, make_db=True, method='cdhit',
#             quiet=False, threads=0, return_just_seq_ids=False, max_memory=800, retries=5, debug=False):
#     '''
#     Perform sequence clustering with CD-HIT.

#     Args:

#         seqs (list or filepath): An iterable of sequences, in any format that ``abutils.utils.sequence.Sequence``
#             can handle. Alternatively, ``seqs`` can be the path to a FASTA-formatted file of sequences. If a path
#             is provided, ``make_db`` is set to ``False`` and the input file will not be deleted following clustering.

#         threshold (float): Clustering identity threshold. Default is ``0.975``.

#         out_file (str): Path to the clustering output file. Default is to use
#             ``tempfile.NamedTemporaryFile`` to generate an output file name.

#         temp_dir (str): Path to the temporary directory. If not provided, ``'/tmp'`` is used.

#         make_db (bool): Whether to build a SQlite database of sequence information. Required
#             if you want to calculate consensus/centroid sequences for the resulting
#             clusters or if you need to access the clustered sequences (not just sequence IDs)
#             Default is ``True``.

#         method (str): Doesn't do anything at this point, stubbed to allow expansion of sequencing tools
#             beyond CD-HIT (UCLUST, etc).
        
#         quiet (bool): If ``True``, surpresses printing of output/progress info. Default is ``False``.

#         threads (int): Number of threads (CPU cores) to be used for clustering. Default is ``0``,
#             which results in all available cores being used.
        
#         return_just_seq_ids (bool): If ``True``, will return a 2D list of sequence IDs
#             (a list containing a list of sequence IDs for each cluster) rather than returning a
#             list of ``Cluster`` objects.

#         max_memory (int): Max memory (in MB) for CD-HIT. Will be passed directly to CD-HIT through
#             the ``-M`` runtime option. Default is ``800``.
        
#         retries (int): Occasionally (and for no discernable reason), the CD-HIT output files are empty. 
#             If this occurs, ``cdhit`` will attempt to retry running CD-HIT until the output files are 
#             not empty. Default is 5.

#         debug (bool): If ``True``, print standard output and standard error from CD-HIT. Default is ``False``.

#     Returns:

#         list: A ``CDHITResult`` object or, if ``return_just_seq_ids`` is ``True``, a 2D list of sequence IDs.
#     '''
#     if make_db:
#         ofile, cfile, seq_db, db_path = cdhit(seqs, out_file=out_file, temp_dir=temp_dir,
#                                               threshold=threshold, make_db=True, quiet=quiet,
#                                               threads=threads, max_memory=max_memory, debug=debug)            
#         return parse_clusters(ofile, cfile, seq_db=seq_db, db_path=db_path, return_just_seq_ids=return_just_seq_ids)
#     else:
#         if isinstance(seqs, (list, tuple)):
#             seqs = [Sequence(s) for s in seqs]
#             seq_dict = {s.id: s for s in seqs}
#         elif os.path.exists(seqs):
#             seq_dict = None
#         ofile, cfile, = cdhit(seqs, out_file=out_file, temp_dir=temp_dir, threads=threads,
#                               threshold=threshold, make_db=False, quiet=quiet,
#                               max_memory=max_memory, debug=debug)
#         return parse_clusters(ofile, cfile, seq_dict=seq_dict, return_just_seq_ids=return_just_seq_ids)


# def cdhit(seqs, out_file=None, temp_dir=None, threshold=0.975, make_db=True,
#           quiet=False, threads=0, max_memory=800, retries=5, debug=False):
#     '''
#     Run CD-HIT.

#     Args:

#         seqs (list or filepath): An iterable of sequences, in any format that `abutils.utils.sequence.Sequence()`
#             can handle. Alternatively, ``seqs`` can be the path to a FASTA-formatted file of sequences. If a file path
#             is provided, ``make_db`` is set to ``False`` and the input file will not be deleted following clustering.

#         threshold (float): Clustering identity threshold. Default is `0.975`.

#         out_file (str): Path to the clustering output file. Default is to use
#             `tempfile.NamedTemporaryFile` to generate an output file name.

#         temp_dir (str): Path to the temporary directory. If not provided, the default 
#             temporary directory is used (typically ``/tmp`` on MacOS and Linux).

#         make_db (bool): Whether to build a SQlite database of sequence information. Required
#             if you want to calculate consensus/centroid sequences for the resulting
#             clusters or if you need to access the clustered sequences (not just sequence IDs). 
#             Default is `True`.
        
#         quiet (bool): If `True`, surpresses printing of output/progress info. Default is `False`.

#         threads (int): Number of threads (CPU cores) to be used for clustering. Default is `0`,
#             which results in all available cores being used.

#         max_memory (int): Max memory (in MB) for CD-HIT. Will be passed directly to CD-HIT through
#             the `-M` runtime option. Default is `800`.

#         retries (int): Occasionally (and for no discernable reason), the CD-HIT output files are empty. 
#             If this occurs, ``cdhit`` will attempt to retry running CD-HIT until the output files are 
#             not empty. Default is 5.

#         debug (bool): If `True`, print standard output and standard error from CD-HIT. Default is `False`.

#     Returns:

#         If `make_db` is `True`, returns the CD-HIT output file path, the CD-HIT cluster file path,
#             a `sqlite3` database connection object, and the database path. If `make_db` is `False`, only the
#             CD-HIT output file path and CD-HIT cluster file path are returned.
#     '''
#     start_time = time.time()
#     delete_input = False if debug else True
#     if not quiet:
#         print('CD-HIT: clustering {} seqeunces'.format(len(seqs)))
#     if out_file is None:
#         out_file = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
#         out_file.close()
#         ofile = out_file.name
#     else:
#         ofile = os.path.expanduser(out_file)
#     cfile = ofile + '.clstr'
#     with open(ofile, 'w') as f: f.write('')
#     with open(cfile, 'w') as f: f.write('')
#     if isinstance(seqs, (list, tuple)):
#         seqs = [Sequence(s) for s in seqs]
#         ifile = _make_cdhit_input(seqs, temp_dir)
#     elif os.path.exists(seqs):
#         ifile = seqs
#         delete_input = False
#     word_size = _get_cdhit_wordsize(threshold)
#     cdhit_cmd = 'cdhit -i {} -o {} -c {} -n {} -d 0 -T {} -M {}'.format(ifile,
#                                                                         ofile,
#                                                                         threshold,
#                                                                         word_size,
#                                                                         threads,
#                                                                         max_memory)
#     while not all([os.path.getsize(cfile), os.path.getsize(cfile)]):
#         cluster = sp.Popen(cdhit_cmd,
#                         shell=True,
#                         stdout=sp.PIPE,
#                         stderr=sp.PIPE)
#         stdout, stderr = cluster.communicate()
#         if not retries:
#             break
#         retries -= 1
#     end_time = time.time()
#     if debug:
#         print(stdout)
#         print(stderr)
#     if delete_input:
#         os.unlink(ifile)
#     if not quiet:
#         print('CD-HIT: clustering took {:.2f} seconds'.format(end_time - start_time))
#     if make_db:
#         if not quiet:
#             print('CD-HIT: building a SQLite3 database')
#         seq_db, db_path = _build_seq_db(seqs, direc=temp_dir)
#         return ofile, cfile, seq_db, db_path
#     return ofile, cfile


# def parse_clusters(out_file, clust_file, seq_db=None, db_path=None, seq_dict=None, return_just_seq_ids=False):
#     '''
#     Parses CD-HIT output.

#     Args:

#         out_file (str): Path to the CD-HIT output file. Required.

#         clust_file (str): Path to the CD-HIT cluster file. Required.

#         seq_db (sqlite.Connection): SQLite3 `Connection` object. Default is ``None``. If not provided and
#             ``return_just_seq_ids`` is False, the returned ``Cluster`` objects will not contain any sequence
#             information beyond the sequence ID.

#         db_path (str): Path to a SQLite3 database file. Default is ``None``. Must be provided if
#             ``seq_db`` is also provided.

#         seq_dict (dict): A ``dict`` mapping sequence IDs to ``abutils.core.sequence.Sequence`` objects. Default
#             is ``None``. Typically used when a relatively small number of sequences are being clustered and
#             creating a ``sqlite3`` database would be overkill.

#         return_just_seq_ids (bool): If ``True``, will return a 2D list of sequence IDs
#             (a list containing a list of sequence IDs for each cluster) rather than returning a
#             list of ``Cluster`` objects.

#     Returns:

#         A ``CDHITResult`` object or, if ``return_just_seq_ids`` is ``True``, a 2D list of sequence IDs.
#     '''
#     raw_clusters = [c.split('\n') for c in open(clust_file, 'r').read().split('\n>')]
#     if return_just_seq_ids:
#         ids = []
#         for rc in raw_clusters:
#             _ids = []
#             for c in rc[1:]:
#                 if c:
#                     _ids.append(c.split()[2][1:-3])
#             ids.append(_ids)
#         os.unlink(out_file)
#         os.unlink(clust_file)
#         return ids
#     os.unlink(out_file)
#     os.unlink(clust_file)
#     clusters = [Cluster(rc, seq_db, db_path, seq_dict) for rc in raw_clusters]
#     return CDHITResult(clusters, seq_db=seq_db, db_path=db_path, seq_dict=seq_dict)


# def _make_cdhit_input(seqs, temp_dir):
#     ifile = tempfile.NamedTemporaryFile(dir=temp_dir, delete=False)
#     ifile.close()
#     fastas = [s.fasta for s in seqs]
#     with open(ifile.name, 'w') as f:
#         f.write('\n'.join(fastas))
#     return ifile.name


# def _get_cdhit_wordsize(threshold):
#     if threshold > 0.7:
#         return 5
#     elif threshold > 0.6:
#         return 4
#     elif threshold > 0.5:
#         return 3
#     return 2


# def _build_seq_db(seqs, direc=None):
#     # '''
#     # Builds a SQLite3 database of sequences.

#     # Inputs are a list of Sequence objects and an optional directory to store the database.
#     # If ::direc:: is not provided, '/tmp' will be used.

#     # Returns a SQLite3 connection object and the database path.
#     # '''
#     direc = direc if direc is not None else '/tmp'
#     db_name = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))
#     db_path = os.path.join(direc, db_name)
#     conn = sqlite3.connect(db_path)
#     c = conn.cursor()
#     create_cmd = '''CREATE TABLE seqs (id text, sequence text)'''
#     insert_cmd = 'INSERT INTO seqs VALUES (?,?)'
#     c.execute('DROP TABLE IF EXISTS seqs')
#     c.execute(create_cmd)
#     c.executemany(insert_cmd, [(str(s.id), str(s.sequence)) for s in seqs])
#     c.execute('CREATE INDEX seq_index ON seqs (id)')
#     return c, db_path



#!/usr/bin/env python
# filename: pl.py


#
# Copyright (c) 2020 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import sys

import pandas as pd

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.patches import Patch

import seaborn as sns

from natsort import natsorted



def feature_kde(data, x, y, hue=None, hue_order=None, colors=None, thresh=0.1,
                show_scatter=True, scatter_size=5, scatter_alpha=0.2,
                fill=False, kde_fill_alpha=0.7, kde_line_alpha=1.0,
                highlight_index=None, highlight_x=None, highlight_y=None, highlight_marker='x',
                highlight_size=90, highlight_color='k', highlight_name=None, highlight_alpha=0.8,
                xlabel=None, ylabel=None, equal_axes=True,
                legend_loc='best', legend_title=None, show_legend_title=True,
                return_ax=False, figsize=[6, 6], figfile=None):
    '''
    Produces a 2-dimensional KDE plot of two features.

    Args:

        data (pd.DataFramne): A pandas dataframe containing the input data.

        x (str): Name of the column in ``df`` containing the feature to be plotted on the x-axis.

        y (str): Name of the column in ``df`` containing the feature to be plotted on the y-axis.

        hue (str): Name of the column in ``df`` containing categories for hue values. For scatter plots, 
                   the categories in ``hue`` will be plotted as different colored points. For KDE plots,
                   ``hue```` categories will each be plotted as differently colored KDE plots
                   on the same plot. 

        hue_order (iterable): Iterable of hue categories, in the order they should be plotted and listed
                              in the legend. If ```hue_order``` contains only a subset of the categories
                              present in ```data[hue]```, only the categories supplied in ```hue_order```
                              will be plotted.

        colors (iterable): List of colors to be used for ```'hue'``` categories. If ```'colors'``` is
                           shorter than the list of hue categories, colors will be reused.

        thresh (float): Threshold for the KDE. Default is ```0.1```.
        
        show_scatter (bool): Show the scatterplot beneath the transparent KDE plot. Default is ```True```.

        scatter_size (int, float): Size of the scatter points. Default is ```5```.

        scatter_alpha (float): Alpha of the scatter points. Default is ```0.2```.

        fill (bool): Fill the KDE plot. Default is ```True```.

        kde_fill_alpha (float): Alpha for the filled KDE plot. If ```fill``` is ```False```,
                                this option is ignored. Default is ```0.7```.
        
        kde_line_alpha (float): Alpha for the KDE plot lines. Default is ```1.0```.

        highlight_index (iterable): An iterabile of index names (present in ```data```) of points
                                    to be highlighted on the KDE plot. If provided, ```highlight_x```
                                    and ```highlight_y``` are ignored.

        highlight_x (iterable): An iterable of x-values for highlighted points. Also requires
                                ```highlight_y```.
        
        highlight_y (iterable): An iterable of y-values for highlighted points. Also requires
                                ```highlight_x```.
        
        highlight_marker (str): The marker style to be used for highlight points. Accepts 
                                standard matplotlib marker styles. Default is ```'x'```. 

        highlight_size (int): Size of the highlight marker. Default is ```90```.

        highlight_color (string or RGB list): Color of the highlight points. Default is black.

        highlight_name (str): Name of the highlights, to be used in the legend. If not supplied,
                              highlight points will not be included in the legend.
        
        highlight_alpha (float): Alpha for the highlight points. Default is ```0.8```.

        xlabel (str): Label for the x-axis. By default, the value for ```x``` is used.

        ylabel (str): Label for the y-axis. By default, the value for ```y``` is used.

        equal_axes (bool): If ```True```, the the limits of the x- and y-axis will be equal.
                           Default is ```True```.
        
        legend_loc (str): Location for the legend. Uses standard matplotlib locations. Default
                          is ```'best'```.

        legend title (str): Title for the legend. By default, ```hue``` is used.

        show_legend_title (bool): Whether or not to show the legend title. Default is ```True```.

        return_ax (bool): If ```True```, return the plot's ```ax``` object. Will not show or save
                          the plot. Default is ```False```.

        figsize (list): A list containg the dimensions of the plot. Default is ```[6, 6]```.

        figfile (str): Path to which the figure will be saved. If not provided, the figure will be
                       shown but not saved to file.
    '''

    # input data
    data = data.copy()
    if x not in data.columns.values:
        print('"{}" is not a column in the supplied dataframe'.format(x))
        return
    if y not in data.columns.values:
        print('"{}" is not a column in the supplied dataframe'.format(y))
        return

    # hue
    if hue is not None:
        if hue_order is None:
            hue_order = natsorted(list(set(data[hue])))
        data = data[data[hue].isin(hue_order)]
    else:
        hue_order = []

    # colors
    n_colors = max(1, len(hue_order))
    if colors is None:
        colors = sns.color_palette(n_colors=n_colors)
        
    plt.figure(figsize=figsize)

    # scatterplots
    if show_scatter:
        if hue_order:
            for h, c in zip(hue_order, colors):
                d = data[data[hue] == h]
                plt.scatter(d[x], d[y], c=[c], s=scatter_size,
                            alpha=scatter_alpha, linewidths=0)
        else:
            plt.scatter(data[x], data[y], c=[colors[0]], s=scatter_size,
                            alpha=scatter_alpha, linewidths=0)

    # kdeplot
    if fill:
        sns.kdeplot(data=data, x=x, y=y, hue=hue, fill=True, alpha=kde_fill_alpha,
                    hue_order=hue_order, palette=colors, thresh=thresh)
    ax = sns.kdeplot(data=data, x=x, y=y, hue=hue, alpha=kde_line_alpha,
                     hue_order=hue_order, palette=colors, thresh=thresh)
    
    # highlighted points
    highlight = any([highlight_index is not None, all([highlight_x is not None, highlight_y is not None])])
    if highlight:
        if highlight_index is not None:
            hidata = data.loc[highlight_index]
            highlight_x = hidata[x]
            highlight_y = hidata[y]
        plt.scatter(highlight_x, highlight_y, zorder=10,
                    s=highlight_size,
                    c=highlight_color,
                    alpha=highlight_alpha,
                    marker=highlight_marker)

    # legend
    legend_labels = hue_order
    if fill:
        handles = []
        for c in colors:
            f = Patch(fc=c, alpha=kde_fill_alpha / 3)
            e = Patch(ec=c, fill=False, lw=1.5)
            handles.append((f, e))
#         handles = [Patch(fc=c, ec=c, alpha=kde_fill_alpha / 3, label=h) for c, h in zip(colors, hue_order)]
    else:
        handles = [Line2D([0], [0], color=c, label=h) for c in colors]
    if highlight_name is not None:
        legend_labels.append(highlight_name)
        handles.append(Line2D([0], [0], marker=highlight_marker, color='w',
                              mec=highlight_color,
                              mfc=highlight_color,
                              ms=highlight_size / 10))
    if show_legend_title:
        legend_title = legend_title if legend_title is not None else hue
    else:
        legend_title = None
    ax.legend(handles, legend_labels, loc=legend_loc, fontsize=13, title=legend_title, frameon=False)
    
    # style the plot
    ax.set_xlabel(xlabel if xlabel is not None else x, fontsize=16)
    ax.set_ylabel(ylabel if ylabel is not None else y, fontsize=16)
    ax.tick_params(axis='both', labelsize=13)

    for spine in ['right', 'top']:
        ax.spines[spine].set_visible(False)
    for spine in ['left', 'bottom']:
        ax.spines[spine].set_position(('outward', 10))

    if equal_axes:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        axlim = [min([xlim[0], ylim[0]]), max([xlim[1], ylim[1]])]
        ax.set_xlim(axlim)
        ax.set_ylim(axlim)
    
    if return_ax:
        return ax
    elif figfile is not None:
        plt.tight_layout()
        plt.savefig(figfile)
    else:
        plt.show()


    
def feature_scatter(data, x, y, hue=None, hue_order=None, colors=None, marker='o', size=20, alpha=0.6,
                    highlight_index=None, highlight_x=None, highlight_y=None, highlight_marker='o',
                    highlight_size=90, highlight_color='k', highlight_name=None, highlight_alpha=0.9,
                    xlabel=None, ylabel=None, equal_axes=True,
                    legend_loc='best', legend_title=None, show_legend_title=True,
                    return_ax=False, figsize=[6, 6], figfile=None):
    '''
    Docstring for feature_scatter.
    '''
    # input data
    data = data.copy()
    if x not in data.columns.values:
        print('"{}" is not a column in the supplied dataframe'.format(x))
        return
    if y not in data.columns.values:
        print('"{}" is not a column in the supplied dataframe'.format(y))
        return
    
    # hue
    if hue is not None:
        if hue_order is None:
            hue_order = natsorted(list(set(data[hue])))
        data = data[data[hue].isin(hue_order)]
    else:
        hue_order = []

    # colors
    n_colors = max(1, len(hue_order))
    if colors is None:
        colors = sns.color_palette(n_colors=n_colors)
        
    plt.figure(figsize=figsize)
    ax = plt.gca()

    # scatterplot
    if hue_order:
        for h, c in zip(hue_order, colors):
            d = data[data[hue] == h]
            plt.scatter(d[x], d[y], c=[c], s=size, marker=marker,
                        alpha=alpha, linewidths=0, label=h)
    else:
        plt.scatter(data[x], data[y], c=[colors[0]], s=size, marker=marker,
                        alpha=alpha, linewidths=0, label=h)

    # highlighted points
    highlight = any([highlight_index is not None, all([highlight_x is not None, highlight_y is not None])])
    if highlight:
        if highlight_index is not None:
            hidata = data.loc[highlight_index]
            highlight_x = hidata[x]
            highlight_y = hidata[y]
        plt.scatter(highlight_x, highlight_y, zorder=10,
                    s=highlight_size,
                    c=highlight_color,
                    alpha=highlight_alpha,
                    marker=highlight_marker,
                    label=highlight_name)
    # legend
    ax.legend(loc=legend_loc, fontsize=13, title=legend_title, frameon=False)
    
    # style the plot
    ax.set_xlabel(xlabel if xlabel is not None else x, fontsize=16)
    ax.set_ylabel(ylabel if ylabel is not None else y, fontsize=16)
    ax.tick_params(axis='both', labelsize=13)

    for spine in ['right', 'top']:
        ax.spines[spine].set_visible(False)
    for spine in ['left', 'bottom']:
        ax.spines[spine].set_position(('outward', 10))

    if equal_axes:
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        axlim = [min([xlim[0], ylim[0]]), max([xlim[1], ylim[1]])]
        ax.set_xlim(axlim)
        ax.set_ylim(axlim)
    
    if return_ax:
        return ax
    elif figfile is not None:
        plt.tight_layout()
        plt.savefig(figfile)
    else:
        plt.show()




# from .core import *
# from .utils import *

import os

from .core import lineage, pair, sequence 

from .utils import alignment, cluster, decorators, jobs, log, phylogeny, pipeline, progbar, s3, seqio, utilities

from .utils import alignment as aln
from .utils import pipeline as path
from .utils import seqio as io

from .version import __version__

BINARY_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'bin'))



#!/usr/bin/env python
# filename: pair.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import csv
import sys
import traceback

from Bio.Seq import Seq

from .sequence import Sequence
from ..utils import germlines
from ..utils.alignment import global_alignment


if sys.version_info[0] > 2:
    STR_TYPES = [str, ]
else:
    STR_TYPES = [str, unicode]


class Pair(object):
    '''
    Holds a pair of sequences, corresponding to HC and LC of a single mAb.

    Input is a list of dicts, with each dict containing sequence information from a single
    chain, formatted as would be returned from a query on a MongoDB database containing
    AbStar output.
    '''
    def __init__(self, seqs, name=None, h_selection_func=None, l_selection_func=None):
        self._seqs = seqs
        self._heavy = None
        self._light = None
        self._heavies = [s for s in seqs if s['chain'] == 'heavy']
        self._lights = [s for s in seqs if s['chain'] in ['kappa', 'lambda']]
        self._name = name
        self._fasta = None
        self._sample = None
        self._subject = None
        self._group = None
        self._experiment = None
        self._timepoint = None
        self._is_pair = None
        self._vrc01_like = None
        self._lineage = None
        self._select_heavy = h_selection_func
        self._select_light = l_selection_func

    def __eq__(self, other):
        return (self.heavy, self.light) == (other.heavy, other.light)

    def __ne__(self, other):
        return not self == other

    def __hash(self):
        return hash((self.heavy, self.light))


    @property
    def heavy(self):
        if self._heavy is None:
            if len(self._heavies) > 0:
                if self._select_heavy is not None:
                    h = self._select_heavy(self._heavies)
                    if all([h is not None, type(h) != Sequence]):
                        h = Sequence(h)
                    self._heavy = h
                else:
                    h = self._heavies[0]
                    if type(h) != Sequence:
                        h = Sequence(h)
                    self._heavy = h
            else:
                self._heavy = None
        return self._heavy

    @heavy.setter
    def heavy(self, heavy):
        self._heavy = heavy

    @property
    def light(self):
        if self._light is None:
            if len(self._lights) > 0:
                if self._select_light is not None:
                    l = self._select_light(self._lights)
                    if all([l is not None, type(l) != Sequence]):
                        l = Sequence(l)
                    self._light = l
                else:
                    l = self._lights[0]
                    if type(l) != Sequence:
                        l = Sequence(l)
                    self._light = l
            else:
                self._light = None
        return self._light

    @light.setter
    def light(self, light):
        self._light = light

    @property
    def is_pair(self):
        if all([self.heavy is not None, self.light is not None]):
            return True
        return False

    @property
    def lineage(self):
        if self._lineage is None:
            self._lineage = self.heavy['clonify']['id']
        return self._lineage

    @property
    def vrc01_like(self):
        if self._vrc01_like is None:
            if any([self.heavy is None, self.light is None]):
                self._vrc01_like = False
            else:
                self._vrc01_like = all([self.heavy['v_gene']['gene'] == 'IGHV1-2', self.light['cdr3_len'] == 5])
        return self._vrc01_like

    @property
    def name(self):
        if self._name is None:
            if self.heavy is not None:
                self._name = self.heavy['seq_id']
            elif self.light is not None:
                self._name = self.light['seq_id']
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def sample(self):
        if self._sample is None:
            slist = []
            if self.experiment is not None:
                slist.append(str(self.experiment))
            if self.group is not None:
                slist.append(str(self.group))
            if self.subject is not None:
                slist.append(str(self.subject))
            if self.timepoint is not None:
                slist.append(str(self.timepoint))
            if slist:
                self._sample = '|'.join(slist)
        return self._sample

    @property
    def subject(self):
        if self._subject is None:
            if self.heavy is not None and 'subject' in self.heavy.keys():
                self._subject = self.heavy['subject']
            elif self.light is not None and 'subject' in self.light.keys():
                self._subject = self.light['subject']
        return self._subject

    @subject.setter
    def subject(self, subject):
        self._subject = subject

    @property
    def group(self):
        if self._group is None:
            if self.heavy is not None and 'group' in self.heavy.keys():
                self._group = self.heavy['group']
            elif self.light is not None and 'group' in self.light.keys():
                self._group = self.light['group']
        return self._group

    @group.setter
    def group(self, group):
        self._group = group

    @property
    def experiment(self):
        if self._experiment is None:
            if self.heavy is not None and 'experiment' in self.heavy.keys():
                self._experiment = self.heavy['experiment']
            elif self.light is not None and 'experiment' in self.light.keys():
                self._experiment = self.light['experiment']
        return self._experiment

    @experiment.setter
    def experiment(self, experiment):
        self._experiment = experiment

    @property
    def timepoint(self):
        if self._timepoint is None:
            if self.heavy is not None and 'timepoint' in self.heavy.keys():
                self._timepoint = self.heavy['timepoint']
            elif self.light is not None and 'timepoint' in self.light.keys():
                self._timepoint = self.light['timepoint']
        return self._timepoint

    @timepoint.setter
    def timepoint(self, timepoint):
        self._timepoint = timepoint


    def refine(self, heavy=True, light=True, species='human'):
        for seq in [s for s in [self.heavy, self.light] if s is not None]:
            try:
                self.remove_ambigs(seq)
                self._refine_v(seq, species)
                self._refine_j(seq, species)
                self._retranslate(seq)
            except:
                print('REFINEMENT FAILED: {}, {} chain'.format(seq['seq_id'], seq['chain']))
                print(traceback.format_exception_only(*sys.exc_info()[:2]))


    @staticmethod
    def remove_ambigs(seq):
        # fix Ns in the nucleotide sequence
        vdj = ''
        for s, g in zip(seq['vdj_nt'], seq['vdj_germ_nt']):
            if s.upper() == 'N':
                vdj += g
            else:
                vdj += s
        seq['vdj_nt'] = vdj
        # fix Xs in the amino acid sequence
        vdj = ''
        for s, g in zip(seq['vdj_aa'], seq['vdj_germ_aa']):
            if s.upper() == 'X':
                vdj += g
            else:
                vdj += s
        seq['vdj_aa'] = vdj

    @staticmethod
    def _refine_v(seq, species):
        '''
        Completes the 5' end of a a truncated sequence with germline nucleotides.
        Input is a MongoDB dict (seq) and the species.
        '''
        vgerm = germlines.get_germline(seq['v_gene']['full'], species)
        aln = global_alignment(seq['vdj_nt'], vgerm)
        prepend = ''
        for s, g in zip(aln.aligned_query, aln.aligned_target):
            if s != '-':
                break
            else:
                prepend += g
        seq['vdj_nt'] = prepend + seq['vdj_nt']

    @staticmethod
    def _refine_j(seq, species):
        '''
        Completes the 3' end of a a truncated sequence with germline nucleotides.
        Input is a MongoDB dict (seq) and the species.
        '''
        jgerm = germlines.get_germline(seq['j_gene']['full'], species)
        aln = global_alignment(seq['vdj_nt'], jgerm)
        append = ''
        for s, g in zip(aln.aligned_query[::-1], aln.aligned_target[::-1]):
            if s != '-':
                break
            else:
                append += g
        seq['vdj_nt'] = seq['vdj_nt'] + append[::-1]

    @staticmethod
    def _retranslate(seq):
        '''
        Retranslates a nucleotide sequence following refinement.
        Input is a Pair sequence (basically a dict of MongoDB output).
        '''
        if len(seq['vdj_nt']) % 3 != 0:
            trunc = len(seq['vdj_nt']) % 3
            seq['vdj_nt'] = seq['vdj_nt'][:-trunc]
        seq['vdj_aa'] = Seq(seq['vdj_nt']).translate()


    def fasta(self, key='vdj_nt', append_chain=True):
        '''
        Returns the sequence pair as a fasta string. If the Pair object contains
        both heavy and light chain sequences, both will be returned as a single string.

        By default, the fasta string contains the 'vdj_nt' sequence for each chain. To change,
        use the <key> option to select an alternate sequence.

        By default, the chain (heavy or light) will be appended to the sequence name:

        >MySequence_heavy

        To just use the pair name (which will result in duplicate sequence names for Pair objects
        with both heavy and light chains), set <append_chain> to False.
        '''
        fastas = []
        for s, chain in [(self.heavy, 'heavy'), (self.light, 'light')]:
            if s is not None:
                c = '_{}'.format(chain) if append_chain else ''
                fastas.append('>{}{}\n{}'.format(s['seq_id'], c, s[key]))
        return '\n'.join(fastas)


def get_pairs(db, collection, experiment=None, subject=None, group=None, name='seq_id',
    delim=None, delim_occurance=1, pairs_only=False, h_selection_func=None, l_selection_func=None):
    '''
    Gets sequences and assigns them to the appropriate mAb pair, based on the sequence name.

    Inputs:

    ::db:: is a pymongo database connection object
    ::collection:: is the collection name, as a string
    If ::subject:: is provided, only sequences with a 'subject' field matching ::subject:: will
        be included. ::subject:: can be either a single subject (as a string) or an iterable
        (list or tuple) of subject strings.
    If ::group:: is provided, only sequences with a 'group' field matching ::group:: will
        be included. ::group:: can be either a single group (as a string) or an iterable
        (list or tuple) of group strings.
    ::name:: is the dict key of the field to be used to group the sequences into pairs.
        Default is 'seq_id'
    ::delim:: is an optional delimiter used to truncate the contents of the ::name:: field.
        Default is None, which results in no name truncation.
    ::delim_occurance:: is the occurance of the delimiter at which to trim. Trimming is performed
        as delim.join(name.split(delim)[:delim_occurance]), so setting delim_occurance to -1 will
        trucate after the last occurance of delim. Default is 1.
    ::pairs_only:: setting to True results in only truly paired sequences (pair.is_pair == True)
        will be returned. Default is False.

    Returns a list of Pair objects, one for each mAb pair.
    '''
    match = {}
    if subject is not None:
        if type(subject) in (list, tuple):
            match['subject'] = {'$in': subject}
        elif type(subject) in STR_TYPES:
            match['subject'] = subject
    if group is not None:
        if type(group) in (list, tuple):
            match['group'] = {'$in': group}
        elif type(group) in STR_TYPES:
            match['group'] = group
    if experiment is not None:
        if type(experiment) in (list, tuple):
            match['experiment'] = {'$in': experiment}
        elif type(experiment) in STR_TYPES:
            match['experiment'] = experiment
    seqs = list(db[collection].find(match))
    return assign_pairs(seqs, name=name, delim=delim,
        delim_occurance=delim_occurance, pairs_only=pairs_only,
        h_selection_func=h_selection_func, l_selection_func=l_selection_func)


def assign_pairs(seqs, name='seq_id', delim=None, delim_occurance=1, pairs_only=False,
                 h_selection_func=None, l_selection_func=None, tenx_annot_file=None):
    '''
    Assigns sequences to the appropriate mAb pair, based on the sequence name.

    Inputs:

    ::seqs:: is a list of dicts, of the format returned by querying a MongoDB containing
        Abstar output.
    ::name:: is the dict key of the field to be used to group the sequences into pairs.
        Default is 'seq_id'
    ::delim:: is an optional delimiter used to truncate the contents of the ::name:: field.
        Default is None, which results in no name truncation.
    ::delim_occurance:: is the occurance of the delimiter at which to trim. Trimming is performed
        as delim.join(name.split(delim)[:delim_occurance]), so setting delim_occurance to -1 will
        trucate after the last occurance of delim. Default is 1.
    ::pairs_only:: setting to True results in only truly paired sequences (pair.is_pair == True)
        will be returned. Default is False.

    Returns a list of Pair objects, one for each mAb pair.
    '''
    # add UMIs to the sequence annotations if a 10xG annotations file is provided
    if tenx_annot_file is not None:
        annots = {}
        with open(tenx_annot_file) as f:
            reader = csv.DictReader(f)
            for r in reader:
                key = 'consensus_id' if 'consensus_id' in r else 'contig_id'
                annots[r[key]] = r['umis']
        for s in seqs:
            s['umis'] = annots.get(s['seq_id'], 0)
        if h_selection_func is None:
            h_selection_func = umi_selector
        if l_selection_func is None:
            l_selection_func = umi_selector
    # identify pairs
    pdict = {}
    for s in seqs:
        if delim is not None:
            pname = delim.join(s[name].split(delim)[:delim_occurance])
        else:
            pname = s[name]
        if pname not in pdict:
            pdict[pname] = [s, ]
        else:
            pdict[pname].append(s)
    pairs = [Pair(pdict[n], name=n,
                  h_selection_func=h_selection_func,
                  l_selection_func=l_selection_func) for n in pdict.keys()]
    if pairs_only:
        pairs = [p for p in pairs if p.is_pair]
    return pairs


def umi_selector(seqs):
    sorted_seqs = sorted(seqs, key=lambda x: x['umis'], reverse=True)
    return sorted_seqs[0]


def deduplicate(pairs, aa=False, ignore_primer_regions=False):
    '''
    Removes duplicate sequences from a list of Pair objects.

    If a Pair has heavy and light chains, both chains must identically match heavy and light chains
    from another Pair to be considered a duplicate. If a Pair has only a single chain,
    identical matches to that chain will cause the single chain Pair to be considered a duplicate,
    even if the comparison Pair has both chains.

    Note that identical sequences are identified by simple string comparison, so sequences of
    different length that are identical over the entirety of the shorter sequence are not
    considered duplicates.

    By default, comparison is made on the nucleotide sequence. To use the amino acid sequence instead,
    set aa=True.
    '''
    nr_pairs = []
    just_pairs = [p for p in pairs if p.is_pair]
    single_chains = [p for p in pairs if not p.is_pair]
    _pairs = just_pairs + single_chains
    for p in _pairs:
        duplicates = []
        for nr in nr_pairs:
            identical = True
            vdj = 'vdj_aa' if aa else 'vdj_nt'
            offset = 4 if aa else 12
            if p.heavy is not None:
                if nr.heavy is None:
                    identical = False
                else:
                    heavy = p.heavy[vdj][offset:-offset] if ignore_primer_regions else p.heavy[vdj]
                    nr_heavy = nr.heavy[vdj][offset:-offset] if ignore_primer_regions else nr.heavy[vdj]
                    if heavy != nr_heavy:
                        identical = False
            if p.light is not None:
                if nr.light is None:
                    identical = False
                else:
                    light = p.light[vdj][offset:-offset] if ignore_primer_regions else p.light[vdj]
                    nr_light = nr.light[vdj][offset:-offset] if ignore_primer_regions else nr.light[vdj]
                    if light != nr_light:
                        identical = False
            duplicates.append(identical)
        if any(duplicates):
            continue
        else:
            nr_pairs.append(p)
    return nr_pairs


def refine(pairs, heavy=True, light=True, species='human'):
    refined_pairs = copy.deepcopy(pairs)
    for p in refined_pairs:
        p.refine(heavy, light, species)
    return refined_pairs



#!/usr/bin/env python
# filename: lineage.py


#
# Copyright (c) 2020 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import colorsys
import random

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns

from natsort import natsorted

from ..utils.color import get_cmap


__all__ = ['lineage_donut', 'annotated_donut']


def lineage_donut(lineages, figfile=None, figsize=(6, 6), pairs_only=False,
                  monochrome_color=None, singleton_color='lightgrey', shuffle_colors=False, seed=1234,
                  text_kws={}, pie_kws={}, fontsize=28, linewidth=2):
    lineages = sorted(lineages, key=lambda x: x.size(pairs_only), reverse=True)
    non_singletons = [l for l in lineages if l.size(pairs_only) > 1]
    singleton_count = sum([1 for l in lineages if l.size(pairs_only) == 1])
    lineage_sizes = [l.size(pairs_only) for l in lineages if l.size(pairs_only) > 1] + [singleton_count]
    if monochrome_color is not None:
        colors = _get_monochrome_colors(monochrome_color, len(non_singletons))
        # we shuffle the colors differently if we're using a monochrome palette, because
        # we want to have the first color (largest lineage) always be the user-supplied
        # monochrome_color. We only want to shuffle the colors starting with the second one.
        if shuffle_colors:
            primary = colors[0]
            secondary = colors[1:]
            random.seed(seed)
            random.shuffle(secondary)
            colors = [primary] + secondary
    else:
        colors = _get_donut_colors(len(non_singletons))
        if shuffle_colors:
            random.seed(seed)
            random.shuffle(colors)
    colors += [singleton_color]

    plt.figure(figsize=figsize)
    ax = plt.gca()
    # fig, ax = plt.subplots()
    ax.axis('equal')
    width = 0.55
    kwargs = dict(colors=colors, startangle=90)
    for k, v in pie_kws.items():
        kwargs[k] = v
    inside, _ = ax.pie(lineage_sizes, radius=1, pctdistance=1 - width / 2, **kwargs)
    plt.setp(inside, width=width, edgecolor='white')

    for w in inside:
        w.set_linewidth(linewidth)

    kwargs = dict(size=fontsize, color='k', va='center', fontweight='bold')
    for k, v in text_kws.items():
        kwargs[k] = v
    ax.text(0, 0, str(sum(lineage_sizes)), ha='center', **kwargs)

    plt.tight_layout()

    if figfile is not None:
        plt.savefig(figfile)
    else:
        plt.show()


def annotated_donut(lineages, annotations, annotation_colors=None, annotation_order=None, annotation_fontsize=20,
                    unannotated_color='grey', unannotated_label='unknown', arc_width=0.125, donut_width=0.425,
                    count_fontsize=48, linewidth=2, arc_linewidth=15, figfile=None, figsize=(8, 8), pairs_only=False,
                    monochrome_color=None, singleton_color='lightgrey', shuffle_colors=False, seed=1234,
                    text_kws={}, pie_kws={}):

    all_sizes = []
    all_colors = []
    if annotation_order is None:
        annotation_order = list(natsorted(annotations.keys()))
    if annotation_colors is None:
        annotation_colors = {a: c for a, c in zip(annotation_order, _get_donut_colors(len(annotation_order)))}

    # annotated_lineages
    for ann in annotation_order:
        alineages = [l for l in lineages if l.name in annotations[ann]]
        non_singletons = [l for l in alineages if l.size(pairs_only) > 1]
        singleton_count = sum([1 for l in alineages if l.size(pairs_only) == 1])
        lineage_sizes = [l.size(pairs_only) for l in alineages if l.size(pairs_only) > 1] + [singleton_count]
        colors = _get_monochrome_colors(annotation_colors[ann], len(non_singletons))
        if shuffle_colors:
            primary = colors[0]
            secondary = colors[1:]
            random.seed(seed)
            random.shuffle(secondary)
            colors = [primary] + secondary
        colors += [singleton_color]
        all_sizes.append(np.array(lineage_sizes))
        all_colors.append(colors)

    # unannotated_lineages
    ulineages = [l for l in lineages if all([l.name not in names for names in annotations.values()])]
    non_singletons = [l for l in ulineages if l.size(pairs_only) > 1]
    singleton_count = sum([1 for l in ulineages if l.size(pairs_only) == 1])
    lineage_sizes = [l.size(pairs_only) for l in ulineages if l.size(pairs_only) > 1] + [singleton_count]
    colors = _get_monochrome_colors(unannotated_color, len(non_singletons))
    if shuffle_colors:
        primary = colors[0]
        secondary = colors[1:]
        random.seed(seed)
        random.shuffle(secondary)
        colors = [primary] + secondary
    colors += [singleton_color]
    all_sizes.append(np.array(lineage_sizes))
    all_colors.append(colors)
    
    # sizes and colors
    sizes = np.array(all_sizes)
    outer_sizes = [s.sum() for s in sizes]
    inner_sizes = [s for sublist in sizes for s in sublist]
    colors = np.array(all_colors)
    outer_colors = [annotation_colors[a] for a in annotation_order] + [unannotated_color]
    inner_colors = [c for sublist in colors for c in sublist]
    
    # make the figure
    plt.subplots(figsize=figsize, subplot_kw=dict(aspect="equal"))
    ax = plt.gca()
    # annotation arc
    wedges, texts = ax.pie(outer_sizes, radius=1, colors=outer_colors, startangle=90,
                           wedgeprops=dict(width=arc_width, edgecolor='w', linewidth=arc_linewidth))
    # donut plot
    ax.pie(inner_sizes, radius=1-arc_width, colors=inner_colors, startangle=90,
           wedgeprops=dict(width=donut_width, edgecolor='w'))
    # sequence count
    kwargs = dict(size=count_fontsize, color='k', va='center', fontweight='bold')
    for k, v in text_kws.items():
        kwargs[k] = v
    ax.text(0, 0, str(sum(inner_sizes)), ha='center', **kwargs)
    
    # add annotation labels
    annot_labels = annotation_order + [unannotated_label]
    kw = dict(arrowprops=dict(arrowstyle="-"),
              fontsize=annotation_fontsize,
              fontweight='medium',
              zorder=0, va="center")
    for i, p in enumerate(wedges):
        ang = (p.theta2 - p.theta1)/2. + p.theta1
        y = np.sin(np.deg2rad(ang))
        x = np.cos(np.deg2rad(ang))
        horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
        connectionstyle = "angle,angleA=0,angleB={}".format(ang)
        kw["arrowprops"].update({"connectionstyle": connectionstyle})
        kw['arrowprops'].update({'color': outer_colors[i]})
        ax.annotate(annot_labels[i], xy=(x, y),
                    xytext=(1.15*np.sign(x), 1.2*y),
                    color=outer_colors[i],
                    horizontalalignment=horizontalalignment, **kw)
    
    # show or save
    plt.tight_layout()
    if figfile is not None:
        plt.savefig(figfile)
    else:
        plt.show()








def _get_donut_colors(N):
    HSV_tuples = [(x * 1.0 / N, 0.8, 0.9) for x in range(N)]
    RGB_tuples = list(map(lambda x: colorsys.hsv_to_rgb(*x), HSV_tuples))[::-1]
    return RGB_tuples


def _get_monochrome_colors(monochrome_color, N):
    cmap = get_cmap(from_color=monochrome_color)
    # this is a bit convoluted, but what's happening is we're getting different colormap
    # values (which range from 0 to 1). Calling cmap(i) returns an rgba tuple, but we just need
    # the rbg, so we drop the a. To make sure that one of the colors isn't pure white,
    # we ask np.linspace() for one more value than we need, reverse the list of RGB tuples
    # so that it goes from dark to light, and drop the lightest value
    RGB_tuples = [cmap(i)[:-1] for i in np.linspace(0, 1, N + 1)][::-1][:-1]
    return RGB_tuples


#!/usr/local/bin/python
# filename: databases.py

#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from abc import ABCMeta, abstractmethod
import os
import sqlite3
import sys
import time

if sys.version_info[0] > 2:
    STR_TYPES = [str, ]
    import pickle
else:
    STR_TYPES = [str, unicode]
    import cPickle as pickle


class SQLiteDatabase():
    """
    base SQLiteDatabase class
    """

    __metaclass__ = ABCMeta
    
    def __init__(self, name=None, direc=None, in_memory=False, table_name=None):
        super(SQLiteDatabase, self).__init__()
        self.name = name
        self.initialized = False
        if all([name is not None, direc is not None]):
            self.dir = os.path.abspath(direc)
            self.path = os.path.join(self.dir, self.name)
            if os.path.isfile(self.path):
                self.initialized = True
        elif in_memory:
            self.dir = None
            self.path = ':memory:'
        self.table_name = 'seqs' if table_name is None else table_name
        self._connection = None
        self._cursor = None
        self._create_table_cmd = None
        if not self.initialized:
            self.create_table()
    
    @property
    @abstractmethod
    def structure(self):
        pass

    @property
    def columns(self):
        return [s[0] for s in self.structure]

    @property
    def insert_cmd(self):
        return 'INSERT INTO {} VALUES ({})'.format(self.table_name,
                ','.join(['?'] * len(self.structure)))

    @property
    def connection(self):
        if self._connection is None:
            self._connection = sqlite3.connect(self.path)
        return self._connection

    @connection.setter
    def connection(self, connection):
        self._connection = connection

    @property
    def cursor(self):
        if self._cursor is None:
            self._cursor = self.connection.cursor()
        return self._cursor

    @cursor.setter
    def cursor(self, cursor):
        self._cursor = cursor

    @property
    def count(self):
        result = self.cursor.execute('SELECT COUNT(*) FROM {}'.format(self.table_name))
        return result.fetchone()[0]

    @property
    def create_table_cmd(self):
        if self._create_table_cmd is None:
            field_string = ', '.join([' '.join(s) for s in self.structure])
            self._create_table_cmd = 'CREATE TABLE {} ({})'.format(self.table_name,
                field_string)
        return self._create_table_cmd
    


    # @abstractmethod
    # def insert_one(self, data, value=None):
    #     '''
    #     Inserts a single database entry
    #     '''
    #     pass

    # @abstractmethod
    # def insert_many(self, data):
    #     '''
    #     Inserts multiple database entries.
    #     '''
    #     pass

    # @abstractmethod
    # def find_one(self):
    #     '''
    #     Returns a single match.
    #     '''
    #     pass

    # @abstractmethod
    # def find_many(self):
    #     '''
    #     Returns multiple matches.
    #     '''
    #     pass

    # @abstractmethod
    # def find_all(self):
    #     '''
    #     Returns all values in a SQLite database.
    #     '''
    #     pass

    # @abstractmethod
    # def delete(self):
    #     '''
    #     Deletes matching database entries
    #     '''
    #     pass

    def commit(self):
        self.connection.commit()

    def close(self):
        self.connection.close()
        del self._cursor
        del self._connection
        self._cursor = None
        self._connection = None

    def create_table(self):
        self.cursor.execute('DROP TABLE IF EXISTS {}'.format(self.table_name))
        self.cursor.execute(self.create_table_cmd)
        self.initialized = True

    def destroy(self):
        self.close()
        if self.path != ':memory:':
            os.unlink(self.path)
    
    # def insert(self, *args, **kwargs):
    #     '''
    #     Alias for ``insert_many``
    #     '''
    #     return self.insert_many(*args, **kwargs)
    
    # def find(self, *args, **kwargs):
    #     '''
    #     Alias for ``find_many``
    #     '''
    #     return self.find_many(*args, **kwargs)

    def index(self, fields):
        '''
        Indexes the database

        Inputs:
          field - the field or list of field on which to create the index.
        '''
        if type(fields) in STR_TYPES:
            fields = [fields]
        index_name = '__'.join(fields) + '__index'
        self.cursor.execute('CREATE INDEX {} ON {} ({})'.format(index_name,
            self.table_name,
            ', '.join(fields)))

    @staticmethod
    def chunker(l, n=900):
        '''
        Yields successive n-sized chunks from l.
        '''
        for i in range(0, len(l), n):
            yield l[i:i + n]




class KeyValueStore(SQLiteDatabase):
    """
    A simple Database with a single key/value per entry. Values are pickled upon insert and
    unpickled upon retrieval, so any pickleable object can be stored. Also provides dictionary-style
    insert and retrieval of values.
    """
    
    def __init__(self, name=None, direc=None, in_memory=False, table_name=None):
        super(KeyValueStore, self).__init__(name=name, direc=direc,
                                               in_memory=in_memory, table_name=table_name)
    
    def __getitem__(self, key):
        return self.find_one(key)

    def __setitem__(self, key, value):
        return self.insert_one(key, value)


    @property
    def structure(self):
        return [('key', 'text'), ('value', 'text')]

    @property
    def insert_cmd(self):
        return 'INSERT INTO {} VALUES ({})'.format(self.table_name,
                ','.join(['?'] * len(self.structure)))


    def keys(self):
        results = self.cursor.execute(
            '''SELECT {0}.key
               FROM {0}'''.format(self.table_name))
        return [r[0] for r in results]


    def values(self):
        results = self.cursor.execute(
            '''SELECT {0}.value
               FROM {0}'''.format(self.table_name))
        return [pickle.loads(r[0]) for r in results]

    
    def items(self):
        results = self.cursor.execute(
            '''SELECT {0}.key, {0}.value
               FROM {0}'''.format(self.table_name))
        return [(r[0], pickle.loads(r[1])) for r in results]


    def insert_one(self, data, value=None):
        '''
        Inserts a single key/value pair

        Inputs:
          data - data to be inserted (as a list or tuple), or optionally, a key name
          value - optional, only used if a key is passed to ``data``
        '''
        if all([value is not None, type(data) in STR_TYPES]):
            data = (data, pickle.dumps(value, protocol=0))
        else:
            data = (data[0], pickle.dumps(data[1], protocol=0))
        with self.connection as conn:
            conn.execute(self.insert_cmd, data)


    def insert_many(self, data):
        '''
        Inserts multiple key/value pairs.

        Inputs:
          data - a list/generator of iterables (lists or tuples) with each iterable containing
                a single key/value pair.
        '''
        # for kv in data:
        #     if len(kv) != len(self.structure):
        #         err = 'Mismatch between one of the supplied values:\n{}\n'.format(kv)
        #         err += 'and the structure of the table:\n{}'.format(self.structure)
        #         raise RuntimeError(err)
        data = ((d[0], pickle.dumps(d[1], protocol=0)) for d in data)
        with self.connection as conn:
            conn.executemany(self.insert_cmd, data)


    def find_one(self, key):
        '''
        Searches a SQLite database.

        Inputs:
          key - a single key (as a string)

        Returns: a single unpickled value
        '''
        query_str = '''SELECT {0}.key, {0}.value
                       FROM {0}
                       WHERE {0}.key LIKE ?'''.format(self.table_name)
        self.cursor.execute(query_str, (key, ))
        try:
            return pickle.loads(self.cursor.fetchone()[1])
        except TypeError:
            return None


    def find_many(self, keys):
        '''
        Searches a SQLite database.

        Inputs:
          keys - a single key (string) or iterable (list/tuple) containing one or more keys

        Returns: a list of unpickled values
        '''
        if type(keys) in STR_TYPES:
            keys = [keys, ]
        results = []
        for chunk in self.chunker(keys):
            result_chunk = self.cursor.execute(
                '''SELECT {0}.key, {0}.value
                   FROM {0}
                   WHERE {0}.key IN ({1})'''.format(self.table_name, ','.join('?' * len(chunk))), chunk)
            results.extend(result_chunk)
        return [pickle.loads(r[1]) for r in results]


    def find_all(self):
        '''
        Returns all values in a SQLite database.

        Inputs:
          keys - a single key (string) or iterable (list/tuple) containing one or more keys

        Returns: a list of unpickled values
        '''
        results = self.cursor.execute(
            '''SELECT {0}.value
               FROM {0}'''.format(self.table_name))
        return [pickle.loads(r[0]) for r in results]


    def delete(self, keys):
        if type(keys) in STR_TYPES:
            keys = [keys, ]
        with self.connection as conn:
            conn.executemany(
                '''DELETE
                   FROM {0}
                   WHERE {0}.key == ?'''.format(self.table_name), keys)


    def index(self, fields='key'):
        '''
        Indexes the database

        Inputs:
          field - the field or list of field on which to create the index.
        '''
        if type(fields) in STR_TYPES:
            fields = [fields]
        index_name = '__'.join(fields) + '__index'
        self.cursor.execute('CREATE INDEX {} ON {} ({})'.format(index_name,
            self.table_name,
            ', '.join(fields)))



#!/usr/bin/env python
# filename: s3.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import os
import subprocess as sp
import sys
import tarfile

from abutils.utils import log

if sys.version_info[0] > 2:
    raw_input = input


def compress_and_upload(data, compressed_file, s3_path, multipart_chunk_size_mb=500,
    method='gz', delete=False, access_key=None, secret_key=None):
    '''
    Compresses data and uploads to S3.

    S3 upload uses ``s3cmd``, so you must either:

        1) Manually configure ``s3cmd`` prior to use (typically using ``s3cmd --configure``).

        2) Configure ``s3cmd`` using ``s3.configure()``.

        3) Pass your access key and secret key to ``compress_and_upload``, which will automatically configure s3cmd.

    .. note:

        ``s3cmd`` configuration only needs to be done once per computer,
        which means that relaunching a cloud instance or Docker image will
        require re-configuration of ``s3cmd``.

    Args:

        data: Can be one of three things:

            1) Path to a single file

            2) Path to a directory

            3) A list of one or more paths to files or directories

        compressed_file (str): Path to the compressed file. Required.

        s3_path (str): The S3 path, with the filename omitted. The S3 filename
          will be the basename of the ``compressed_file``. For example::

            compress_and_upload(data='/path/to/data',
                                compressed_file='/path/to/compressed.tar.gz',
                                s3_path='s3://my_bucket/path/to/')

          will result in an uploaded S3 path of ``s3://my_bucket/path/to/compressed.tar.gz``

        method (str): Compression method. Options are ``'gz'`` (gzip) or ``'bz2'`` (bzip2).
            Default is ``'gz'``.

        delete (bool): If ``True``, the ``compressed_file`` will be deleted after upload
            to S3. Default is ``False``.

        access_key (str): AWS access key.

        secret_key (str): AWS secret key.
    '''
    logger = log.get_logger('s3')
    if all([access_key, secret_key]):
        configure(access_key=access_key, secret_key=secret_key, logger=logger)
    compress(data, compressed_file, fmt=method, logger=logger)
    put(compressed_file, s3_path, multipart_chunk_size_mb=multipart_chunk_size_mb, logger=logger)
    if delete:
        os.unlink(compressed_file)


def put(f, s3_path, multipart_chunk_size_mb=500, logger=None):
    '''
    Uploads a single file to S3, using s3cmd.

    Args:

        f (str): Path to a single file.

        s3_path (str): The S3 path, with the filename omitted. The S3 filename
            will be the basename of the ``f``. For example::

                put(f='/path/to/myfile.tar.gz', s3_path='s3://my_bucket/path/to/')

            will result in an uploaded S3 path of ``s3://my_bucket/path/to/myfile.tar.gz``
    '''
    if not logger:
        logger = log.get_logger('s3')
    fname = os.path.basename(f)
    target = os.path.join(s3_path, fname)
    s3cmd_cline = 's3cmd put {} {} --multipart-chunk-size-mb {}'.format(f,
                                                                        target,
                                                                        multipart_chunk_size_mb)
    print_put_info(fname, target, logger)
    s3cmd = sp.Popen(s3cmd_cline,
                     stdout=sp.PIPE,
                     stderr=sp.PIPE,
                     shell=True)
    stdout, stderr = s3cmd.communicate()


def print_put_info(fname, target, logger):
    logger.info('')
    logger.info('')
    logger.info('')
    logger.info('-' * 25)
    logger.info('UPLOADING TO S3')
    logger.info('-' * 25)
    logger.info('')
    logger.info('File: {}'.format(fname))
    logger.info('Target S3 location: {}'.format(target))



def compress(d, output, fmt='gz', logger=None):
    '''
    Creates a compressed/uncompressed tar file.

    Args:

        d: Can be one of three things:

            1. the path to a single file, as a string

            2. the path to a single directory, as a string

            3. an iterable of file or directory paths

        output (str): Output file path.

        fmt: Compression method. Options are ``'gz'`` (gzip),
            ``'bz2'`` (bzip2) and ``'none'`` (uncompressed). Default is ``'gz'``.
    '''
    if not logger:
        logger = log.get_logger('s3')
    if type(d) not in [list, tuple]:
        d = [d, ]
    d = [os.path.expanduser(_d) for _d in d]
    print_compress_info(d, output, compress, logger)
    if fmt.lower() == 'none':
        fmt = ''
    elif fmt.lower() not in ['gz', 'bz2']:
        logger.info('Compression option ("{}") is invalid.\nFalling back to uncompressed.'.format(fmt))
        fmt = ''
    output = os.path.expanduser(output)
    tar = tarfile.open(output, 'w:{}'.format(fmt))
    for obj in d:
        tar.add(obj)
    tar.close()
    return output


def print_compress_info(d, output, compress, logger):
    if not logger:
        logger = log.get_logger('s3')
    dirs = [obj for obj in d if os.path.isdir(obj)]
    files = [obj for obj in d if os.path.isfile(obj)]
    logger.info('')
    logger.info('')
    logger.info('')
    logger.info('-' * 25)
    logger.info('COMPRESSING DATA')
    logger.info('-' * 25)
    logger.info('')
    logger.info('Ouptut file: {}'.format(output))
    logger.info('Compression: {}'.format(compress.lower()))
    if dirs:
        d = 'directories' if len(dirs) > 1 else 'directory'
        logger.info('Found {} {} to compress: {}'.format(len(dirs), d,
                                                         ', '.join(dirs)))
    if files:
        f = 'files' if len(files) > 1 else 'file'
        logger.info('Found {} {} to compress: {}'.format(len(files), f,
                                                         ', '.join(files)))


def configure(access_key=None, secret_key=None, logger=None):
    '''
    Configures s3cmd prior to first use.

    If no arguments are provided, you will be prompted to enter
    the access key and secret key interactively.

    Args:

        access_key (str): AWS access key

        secret_key (str): AWS secret key
    '''
    if not logger:
        logger = log.get_logger('s3')
    if not all([access_key, secret_key]):
        logger.info('')
        access_key = input('AWS Access Key: ')
        secret_key = input('AWS Secret Key: ')
    _write_config(access_key, secret_key)
    logger.info('')
    logger.info('Completed writing S3 config file.')
    logger.info('')


def _write_config(access_key, secret_key):
    cfg_string = '[default]\n'
    cfg_string += 'access_key = {}\n'.format(access_key)
    cfg_string += 'secret_key = {}\n'.format(secret_key)
    cfg_string += CONFIG_DEFAULTS
    cfg_file = os.path.expanduser('~/.s3cfg')
    open(cfg_file, 'w').write(cfg_string)


CONFIG_DEFAULTS = '''
access_token =
add_encoding_exts =
add_headers =
bucket_location = US
cache_file =
cloudfront_host = cloudfront.amazonaws.com
default_mime_type = binary/octet-stream
delay_updates = False
delete_after = False
delete_after_fetch = False
delete_removed = False
dry_run = False
enable_multipart = True
encoding = UTF-8
encrypt = False
expiry_date =
expiry_days =
expiry_prefix =
follow_symlinks = False
force = False
get_continue = False
gpg_command = /usr/bin/gpg
gpg_decrypt = %(gpg_command)s -d --verbose --no-use-agent --batch --yes --passphrase-fd %(passphrase_fd)s -o %(output_file)s %(input_file)s
gpg_encrypt = %(gpg_command)s -c --verbose --no-use-agent --batch --yes --passphrase-fd %(passphrase_fd)s -o %(output_file)s %(input_file)s
gpg_passphrase =
guess_mime_type = True
host_base = s3.amazonaws.com
host_bucket = %(bucket)s.s3.amazonaws.com
human_readable_sizes = False
ignore_failed_copy = False
invalidate_default_index_on_cf = False
invalidate_default_index_root_on_cf = True
invalidate_on_cf = False
list_md5 = False
log_target_prefix =
max_delete = -1
mime_type =
multipart_chunk_size_mb = 250
preserve_attrs = True
progress_meter = True
proxy_host =
proxy_port = 0
put_continue = False
recursive = False
recv_chunk = 4096
reduced_redundancy = False
restore_days = 1
send_chunk = 4096
server_side_encryption = False
simpledb_host = sdb.amazonaws.com
skip_existing = False
socket_timeout = 300
urlencoding_mode = normal
use_https = True
use_mime_magic = True
verbosity = WARNING
website_endpoint = http://%(bucket)s.s3-website-% (location)s.amazonaws.com/
website_error =
website_index = index.html
'''



#!/usr/bin/env python
# filename: features.py


#
# Copyright (c) 2020 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


__all__ = ['assign_cellhash_groups', 'classify_features',
           'positive_feature_cutoff', 'negative_feature_cutoff']


import numpy as np
import pandas as pd

import matplotlib.pyplot as plt

from sklearn.neighbors import KernelDensity

from scipy.signal import argrelextrema
from scipy.stats import scoreatpercentile


def assign_cellhash_groups(df, hashnames=None, rename_dict=None, log_normalize=True,
                           threshold_minimum=4.0, threshold_maximum=10.0, kde_maximum=15.0, 
                           assignments_only=False, return_intermediates=False, debug=False):
    hash_df = df.copy()
    # make sure the dataframe is counts + 1
    if 0 in hash_df.values:
        hash_df += 1
    if log_normalize:
        hash_df = hash_df.applymap(np.log2)
    if hashnames is None:
        hashnames = hash_df.columns.values
    if rename_dict is None:
        rename_dict = {h: h for h in hashnames}

    thresholds = {}
    for hashname in hashnames:
        if debug:
            print(hashname)
        thresholds[hashname] = positive_feature_cutoff(hash_df[hashname],
                                                       threshold_minimum=threshold_minimum,
                                                       threshold_maximum=threshold_maximum,
                                                       kde_maximum=kde_maximum,
                                                       debug=debug)
    
    if debug:
        print('THRESHOLDS')
        print('----------')
        for hashname in hashnames:
            print(f'{hashname}: {thresholds[hashname]}')

    assignments = []
    for i, row in hash_df[hashnames].iterrows():
        a = [h for h in hashnames if row[h] >= thresholds[h]]
        if len(a) == 1:
            assignment = rename_dict.get(a[0], 'not found')
        elif len(a) > 1:
            assignment = 'doublet'
        else:
            assignment = 'unassigned'
        assignments.append(assignment)
    hash_df['assignment'] = assignments
    
    if assignments_only:
        return hash_df['assignment']
    else:
        if return_intermediates:
            return hash_df
        else:
            return hash_df[hashnames + ['assignment']]


def classify_features(df, positive=None, negative=None,
                      positive_threshold_maximum=10.0, positive_threshold_minimum=4.0, positive_kde_maximum=15.0,
                      negative_threshold_maximum=10.0, negative_threshold_minimum=4.0, negative_kde_maximum=15.0,
                      negative_threshold_denominator=2, classifications_only=False, debug=False):
    '''
    Docstring for classify_features.
    '''
    df = df.copy()
    data = {}
    
    if positive is None:
        positive = []
    if negative is None:
        negative = []
    
    for p in positive:
        if debug:
            print(p)
        cutoff = positive_feature_cutoff(df[p],
                                         threshold_maximum=positive_threshold_maximum,
                                         threshold_minimum=positive_threshold_minimum,
                                         kde_maximum = positive_kde_maximum,
                                         debug=debug)
        data[p] = df[p] >= cutoff
    for n in negative:
        if debug:
            print(n)
        cutoff = negative_feature_cutoff(df[n],
                                         threshold_maximum=negative_threshold_maximum,
                                         threshold_minimum=negative_threshold_minimum,
                                         kde_maximum = negative_kde_maximum,
                                         denominator=negative_threshold_denominator,
                                         debug=debug)
        data[n] = df[n] > cutoff
    
    if classifications_only:
        return pd.DataFrame(data)
    else:
        data = {'{}_classified'.format(k): v for k, v in data.items()}
        return df.join(pd.DataFrame(data))
        
    


def positive_feature_cutoff(vals, threshold_maximum=10.0, threshold_minimum=4.0,
                            kde_maximum=15.0, debug=False):
    a = np.array(vals)
    k = _bw_silverman(a)
    kde = KernelDensity(kernel='gaussian', bandwidth=k).fit(a.reshape(-1, 1))
    s = np.linspace(0, kde_maximum, num=int(kde_maximum * 100))
    e = kde.score_samples(s.reshape(-1,1))
    
    all_min, all_max = argrelextrema(e, np.less)[0], argrelextrema(e, np.greater)[0]
    if len(all_min) > 1:
        _all_min = np.array([m for m in all_min if s[m] <= threshold_maximum and s[m] >= threshold_minimum])
        min_vals = zip(_all_min, e[_all_min])
        mi = sorted(min_vals, key=lambda x: x[1])[0][0]
        cutoff = s[mi]
    elif len(all_min) == 1:
        mi = all_min[0]
        cutoff = s[mi]
    else:
        cutoff = None

    if debug:
        plt.plot(s, e)
        plt.show()
        print('bandwidth: {}'.format(k))
        print('local minima: {}'.format(s[all_min]))
        print('local maxima: {}'.format(s[all_max]))
        if cutoff is not None:
            print('cutoff: {}'.format(cutoff))
        else:
            print('WARNING: no local minima were found, so the threshold could not be calculated.')
        print('\n\n')

    return cutoff


def negative_feature_cutoff(vals, threshold_maximum=10.0, threshold_minimum=4.0,
                            kde_maximum=15.0, denominator=2, debug=False):
    a = np.array(vals)
    k = _bw_silverman(a)
    kde = KernelDensity(kernel='gaussian', bandwidth=k).fit(a.reshape(-1, 1))
    s = np.linspace(0, kde_maximum, num=int(kde_maximum * 100))
    e = kde.score_samples(s.reshape(-1,1))
    
    all_min, all_max = argrelextrema(e, np.less)[0], argrelextrema(e, np.greater)[0]
    if len(all_min) > 1:
        _all_min = np.array([m for m in all_min if s[m] <= threshold_maximum and s[m] >= threshold_minimum])
        min_vals = zip(_all_min, e[_all_min])
        mi = sorted(min_vals, key=lambda x: x[1])[0][0]
        i = list(all_min).index(mi)
        try:
            if s[mi] > s[all_max[i]]:
                ma = all_max[i]
            else:
                ma = all_max[i - 1]
        except IndexError:
            ma = all_max[i-1]
        cutoff = (s[mi] + s[ma]) / denominator
    elif len(all_min) == 1:
        mi = all_min[0]
        ma = all_max[0]
        cutoff = (s[mi] + s[ma]) / denominator
    else:
        cutoff = None

    if debug:
        plt.plot(s, e)
        plt.show()
        print('bandwidth: {}'.format(k))
        print('local minima: {}'.format(s[all_min]))
        print('local maxima: {}'.format(s[all_max]))
        if cutoff is not None:
            print('cutoff: {}'.format(cutoff))
        else:
            print('WARNING: no local minima were found, so the threshold could not be calculated.')
        print('\n\n')

    return cutoff
    
    
def _bw_silverman(x):
    normalize = 1.349
    IQR = (scoreatpercentile(x, 75) - scoreatpercentile(x, 25)) / normalize
    std_dev = np.std(x, axis=0, ddof=1)
    if IQR > 0:
        A = np.minimum(std_dev, IQR)
    else:
        A = std_dev
    n = len(x)
    return .9 * A * n ** (-0.2)







#!/usr/bin/env python
# filename: progbar.py


#
# Copyright (c) 2015 Bryan Briney
# License: The MIT license (http://opensource.org/licenses/MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or
# substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


from __future__ import absolute_import, division, print_function, unicode_literals

import sys
from datetime import datetime


def progress_bar(finished, total, start_time=None, extra_info=None,
                 autocomplete=True, complete=False, completion_string='\n'):
    '''
    Prints an ASCII progress bar.

    Each call to ``progress_bar`` will update the progress bar. An example
    of tracking the progress of a list of items would look like::

        job_list = [job1, job2, job3, ... jobN]
        total_jobs = len(job_list)

        #initialize the progress bar
        progress_bar(0, total_jobs)

        # do the jobs
        for i, job in enumerate(job_list):
            do_job(job)
            progress_bar(i + 1, total_jobs)

    Args:

        finished (int): Number of finished jobs.

        total (int): Total number of jobs.

        start_time (datetime): Start time, as a ``datetime.datetime`` object.
            Only required if you want to display execution time alongside
            the progress bar. If not provided, execution time is not shown.
        
        extra_info (str): A string containing extra information to be displayed
            at the end of the progbar string. Examples include the number of failed
            jobs, the name of the job batch currently being processed, etc.

        complete (bool): If `True`, will append `completion_string` to the end
            of the progbar string.

        completion_string (str): Will be appended to the progbar string if
            `complete` is `True`. Default is `'\n'`.

    '''
    pct = int(100. * finished / total)
    ticks = int(pct / 2)
    spaces = int(50 - ticks)
    if start_time is not None:
        elapsed = (datetime.now() - start_time).seconds
        minutes = int(elapsed / 60)
        seconds = int(elapsed % 60)
        minute_str = '0' * (2 - len(str(minutes))) + str(minutes)
        second_str = '0' * (2 - len(str(seconds))) + str(seconds)
        prog_bar = '\r({}/{}) |{}{}|  {}%  ({}:{})  '.format(finished, total,
            '|' * ticks, ' ' * spaces, pct, minute_str, second_str)
    else:
        prog_bar = '\r({}/{}) |{}{}|  {}%  '.format(finished, total,
            '|' * ticks, ' ' * spaces, pct)
    if extra_info is not None:
        prog_bar += str(extra_info)
    if autocomplete and finished == total:
        prog_bar += completion_string
    if complete:
        prog_bar += completion_string
    sys.stdout.write(prog_bar)
    sys.stdout.flush()
